##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IApplication.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################

from GangaCore.GPIDev.Adapters.IApplication import IApplication
from GangaCore.Core.GangaRepository import getRegistry
from GangaCore.GPIDev.Base.Proxy import GPIProxyObjectFactory, isType
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
from GangaCore.GPIDev.Lib.File.File import File, ShareDir
import os
import shutil
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem

import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()


from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.files import expandfilename

config = getConfig("Preparable")

class IPrepareApp(IApplication):

    """
    Base class for all applications which can be placed into a prepared\
    state. 
    """
    _schema = Schema(Version(0, 0), {'hash': SimpleItem(defvalue=None, typelist=[None, str], hidden=1)})
    _category = 'applications'
    _name = 'PrepareApp'
    _hidden = 1

    __slots__ = list()

    def _auto__init__(self, unprepare=None):
        """
        Function called when initializing from the Proxy layer i.e. interactive prompt or 'import ganga'
        Args:
            unprepare (bool): a parameter which unprepares an app when it's created new i.e. don't copy prepared sandboxes
        """
        if unprepare is True:
            logger.debug("Calling unprepare() from IPrepareApp's _auto__init__()")
            self.unprepare()

    def prepare(self, force=False):
        """
        Base class for all applications which can be placed into a prepared\
        state. 
        Args:
            force (bool) : forces the prepare function to be called no matter what when True
        """
        pass

    def post_prepare(self):
        """
        Put any methods that should always be run at the end of the preparation process here.
        """
        self.calc_hash()

    def unprepare(self, force=False):
        """
        Revert an application back to the exact state it was in prior to being\
        prepared.
        Args:
            force (bool): causes unprepare to run always or not if True
        """
        logger.debug("Running unprepare() from IPrepareApp")
        if self.is_prepared is True:
            self.is_prepared = None
        elif self.is_prepared is not None:
            self.is_prepared = None
        self.hash = None

    def getSharedPath(self):
        """
        Return the full path of the shared directory where files are placed
        """
        share_folder = os.path.join(expandfilename(getConfig('Configuration')['gangadir']), 'shared', getConfig('Configuration')['user'])
        if isinstance(self.is_prepared, ShareDir):
            return os.path.join(share_folder, self.is_prepared.name)
        else:
            return share_folder

    def copyIntoPrepDir(self, obj2copy):
        """
        Method for actually copying the "obj2copy" object to the prepared state dir of this application
        Args:
            obj2copy (bool): is a string (local) address of a file to be copied as it's passed to shutil.copy2
        """
        shr_dir = self.getSharedPath()

        if not os.path.isdir(shr_dir):
            os.makedirs(shr_dir)
        shutil.copy2(obj2copy, shr_dir)
        logger.debug("Copying %s into: %s" % (obj2copy, shr_dir))

    def copyPreparables(self):
        """
        This method iterates over all attributes in an application and decides\
        whether they should be persisted (i.e. copied) in the Shared Directory\
        when the application is prepared.
        If an IOError is raised when attempting the copy, this method returns 0\
        otherwise it returns 1.
        """
        send_to_sharedir = []
        for name, item in self._schema.allItems():
            if item['preparable']:
                logger.debug('Found preparable %s' % (name))
                logger.debug('adding to sharedir %s' % (self.__getattribute__(name)))
                send_to_sharedir.append(self.__getattribute__(name))

        for prepitem in send_to_sharedir:
            logger.debug('working on %s' % (prepitem))
            # we may have a list of files/strings
            if isType(prepitem, (list, GangaList)):
                logger.debug('found a list')
                for subitem in prepitem:
                    if isType(subitem, str):
                        logger.debug('found a string')
                    elif isType(subitem, File) and subitem.name != '':
                        logger.debug('found a file')
                        logger.debug('Sending file object %s to shared directory' % subitem.name)
                        try:
                            self.copyIntoPrepDir(subitem.name)
                        except IOError as e:
                            logger.error(e)
                            return 0
            elif isinstance(prepitem, str):
                logger.debug('found a string')
            elif isType(prepitem, File) and prepitem.name != '':
                logger.debug('found a file')
                logger.debug('Sending "File" object %s to shared directory' % prepitem.name)
                try:
                    self.copyIntoPrepDir(prepitem.name)
                except IOError as e:
                    logger.error(e)
                    return 0
            else:
                logger.debug('Nothing worth copying found in %s' % (prepitem))
        return 1

    def calc_hash(self, verify=False):
        """Calculate the MD5 digest of the application's preparable attribute(s), and store
        that value in the application schema. The value is recalculated (and compared against
        the initial value) every time the application is written to the Ganga repository. This
        allows warnings to be generated should an application's locked attributes be changed 
        post-preparation.
        Args:
            verify (bool) : If the hash is to be verified in the future True save it to the hash schema attribute
        """
        from GangaCore.GPIDev.Base.Proxy import runProxyMethod
        import io
        try:
            import hashlib
            digest = hashlib.new('md5')
        except Exception as err:
            logger.debug("Err: %s" % err)
            import md5
            digest = md5.new()

        sio = io.StringIO()
        runProxyMethod(self, 'printTree', sio)
        digest.update(sio.getvalue().encode('utf-8'))
        tmp = sio.getvalue()
        if verify == False:
            self.hash = digest.hexdigest()
        else:
            # we return true if this is called with verify=True and the current hash is the same as that stored in the schema.
            # this is checked immediately prior to (re)writing the object to
            # the repository
            return digest.hexdigest() == self.hash

    #printTree is only ever run on applications, from within IPrepareApp.py
    #if you (manually) try to run printTree on anything other than an application, it will not work as expected
    #see the relevant code in VPrinter to understand why
    def printTree(self, f=None, sel='preparable' ):
        ## After fixing some bugs we are left with incompatible job hashes. This should be addressd before removing
        ## This particular class!
        from GangaCore.GPIDev.Base.VPrinterOld import VPrinterOld
        self.accept(VPrinterOld(f, sel))

    def incrementShareCounter(self, shared_directory):
        """
        Function which is used to increment the number of (sub)jobs which share the prepared sandbox
        managed by this app
        Args:
            shared_directory (ShareDir): The ShareDir object managed by this app
        """
        logger.debug('Incrementing shared directory reference counter')
        shareref = getRegistry("prep").getShareRef()
        logger.debug('within incrementShareCounter, calling increase')
        shareref.increase(shared_directory)

    def decrementShareCounter(self, shared_directory, remove=0):
        """
        Function which is used to decrement the number of (sub)jobs which share the prepared sandbox
        managed by this app
        Args:
            shared_directory (ShareDir): The ShareDir object managed by this app
        """
        remove = remove
        logger.debug('Decrementing shared directory reference counter')
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        shareref.decrease(shared_directory, remove)

    def getShareCounterVal(self, shared_directory):
        """
        Function to get the current value of the counter
        Args:
            shared_directory_name (str): full name of directory managed by this app
        """
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        return shareref.counterVal(shared_directory)

    def listShareDirs(self):
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        shareref

    def listShareDirContents(self, shared_directory_name):
        """
        Function which is used to list the contents of the prepared sandbox folder managed by this app
        Args:
            shared_directory_name (str): full name of directory managed by this app
        """
        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        shareref.ls(shared_directory_name)

    def checkPreparedHasParent(self, prepared_object):
        """
        Function which is used to check if a prepared app has a parent in a registry 
        Args:
            prepared_object (IPrepareApp): object in a registry which manages a prepared state
        """
        if prepared_object._getRegistry() is None:
            shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
            shareref.registerForRemoval(prepared_object.is_prepared.name)
            logger.info('Application is not currently associated with a persisted Ganga object')
            logger.info('(e.g. box, job, task). Both the prepared application and the contents of')
            logger.info('its shared directory will be lost when Ganga exits.')
            logger.info('Shared directory location: %s' % (self.is_prepared.name))
            # logger.error(self.listShareDirContents(prepared_object.is_prepared.name))
        else:
            self.incrementShareCounter(prepared_object.is_prepared)




# Import classes that should be in the Tasks namespace
## Import the list of tasks and the task and abstract job definition

## Tasks
from NA62Task import NA62Task
from NA62Transform import NA62Transform
from NA62Unit import NA62Unit




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
import os.path
import pprint
import tempfile
from GangaCore.Utility.Shell import Shell
from GangaCore.Utility.files import fullpath
from GangaCore.Utility.util import unique
import GangaCore.Utility.logging
import GangaCore.Utility.Config
from GangaCore.Core.exceptions import GangaException
from . import CMTUtils
from . import cmakeUtils
import subprocess
import time
import collections

logger = GangaCore.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

def make(self, arguments):

    if not self.newStyleApp is True:
        return CMTUtils.make(arguments)
    else:
        return cmakeUtils.make(arguments)

def get_user_platform(self, env=os.environ):

    if not self.newStyleApp is True:
        return CMTUtils.get_user_platform(env)
    else:
        return cmakeUtils.get_user_platform(env)


def update_project_path(self, user_release_area, env=os.environ):

    if not self.newStyleApp is True:
        return CMTUtils.update_project_path(user_release_area, env)
    else:
        return cmakeUtils.update_project_path(user_release_area, env)

def get_user_dlls(self, appname, version, user_release_area, platform, env):

    if not self.newStyleApp is True:
        return CMTUtils.get_user_dlls(appname, version, user_release_area, platform, env)
    else:
        return cmakeUtils.get_user_dlls(appname, version, user_release_area, platform, env)

def pyFileCollector(dir, file_list, subdir_dict, depth_cut, depth=0, zerodepth_pathlength=0):
    if zerodepth_pathlength == 0:
        zerodepth_pathlength = len(dir) + 1
    sub_pys = []
    for item in os.listdir(dir):
        file_path = os.path.join(dir, item)
        if (file_path.endswith('.py')):
            if os.path.exists(file_path):
                if depth == 0:
                    file_list.append(file_path)
                else:
                    sub_pys.append(file_path)
            else:
                logger.warning(
                    "File %s in %s does not exist. Skipping...", str(item), str(dir))
        elif os.path.isdir(file_path) and not os.path.islink(file_path):
            if depth >= depth_cut:
                continue
            pyFileCollector(
                file_path, file_list, subdir_dict, depth_cut, depth + 1, zerodepth_pathlength)

    if depth != 0:
        subdir_dict[dir[zerodepth_pathlength:]] = sub_pys


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

def shellEnv_cmd(cmd, environ=None, cwdir=None):
    pipe = subprocess.Popen(cmd,
                            shell=True,
                            env=environ,
                            cwd=cwdir,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=subprocess.DEVNULL)
    stdout, stderr = pipe.communicate()
    while pipe.poll() is None:
        time.sleep(0.5)
    return pipe.returncode, stdout, stderr


def shellEnvUpdate_cmd(cmd, environ=os.environ, cwdir=None):
    import tempfile
    import pickle
    f = tempfile.NamedTemporaryFile(mode='w+b')
    fname = f.name
    f.close()

    if not cmd.endswith(';'):
        cmd += ';'
    envdump = 'import os, pickle;'
    envdump += 'f=open(\'%s\',\'w+b\');' % fname
    envdump += 'pickle.dump(os.environ,f);'
    envdump += 'f.close();'
    envdumpcommand = 'python -c \"%s\"' % envdump
    cmd += envdumpcommand

    pipe = subprocess.Popen(cmd,
                            shell=True,
                            env=environ,
                            cwd=cwdir,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=subprocess.DEVNULL)
    stdout, stderr = pipe.communicate()
    while pipe.poll() is None:
        time.sleep(0.5)

    f = open(fname, 'r+b')
    environ = environ.update(pickle.load(f))
    f.close()
    os.system('rm -f %s' % fname)

    return pipe.returncode, stdout, stderr

    #\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
    # NEW, better version?

CommandOutput = collections.namedtuple(
    'CommandOutput', ['returncode', 'stdout', 'stderr'])


class TimeoutException(Exception):

    def __init__(self, message):
        super(TimeoutException, self).__init__(message)


def run(cmd, env=None, cwd=None, timeout=None):
    proc = subprocess.Popen(cmd,
                            shell=True,
                            env=env,
                            cwd=cwd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=subprocess.DEVNULL)
    if timeout is not None:
        time_start = time.time()
        while proc.poll() is None:
            if time.time() - time_start >= timeout:
                proc.kill()
                raise TimeoutException("Command '%s' timed out!" % cmd)
            time.sleep(0.5)
    stdout, stderr = proc.communicate()
    return CommandOutput(proc.returncode, stdout, stderr)


def runUpdate(cmd, env=None, cwd=None, timeout=None):
    (fd, filename) = tempfile.mkstemp()
    command_output = run(cmd + '; printenv &> ' + filename, env, cwd, timeout)
    with os.fdopen(fd, 'r') as file:
        env.update(dict([tuple(line.split('=', 1))
                         for line in file.read().splitlines()]))
    return command_output


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def fillPackedSandbox(sandbox_files, destination):
    """Put all sandbox_files into tarball called name and write it into to the destination.
    Arguments:
    'sandbox_files': a list of File or FileBuffer objects.
    'destination': a string representing the destination filename
    Return: a list containing a path to the tarball
    """
    if not sandbox_files:
        return []

    # Generalised version from Ganga/Core/Sandbox/Sandbox.py

    import tarfile
    import stat
    #tf = tarfile.open(destination,"w:gz")

    # "a" = append with no compression
    # creates file if doesn't exist
    # cant append to a compressed tar archive so must compress later
    dir, filename = os.path.split(destination)
    if not os.path.isdir(dir):
        os.makedirs(dir)
    tf = tarfile.open(destination, "a")
    tf.dereference = True  # --not needed in Windows

    for f in sandbox_files:
        try:
            contents = f.getContents()   # is it FileBuffer?

        except AttributeError:         # File
            try:
                fileobj = open(f.name, 'wb')
            except Exception as err:
                raise GangaException("File %s does not exist." % f.name)
            tinfo = tf.gettarinfo(
                f.name, os.path.join(f.subdir, os.path.basename(f.name)))

        else:                          # FileBuffer
            from io import StringIO, BytesIO
            fileobj = BytesIO(contents)

            tinfo = tarfile.TarInfo()
            tinfo.name = os.path.join(f.subdir, os.path.basename(f.name))
            import time
            tinfo.mtime = time.time()
            tinfo.size = len(fileobj.getvalue())

        if f.isExecutable():
            tinfo.mode = tinfo.mode | stat.S_IXUSR
        tf.addfile(tinfo, fileobj)

    tf.close()

    return [destination]


def gzipFile(filename, outputfilename=None, removeOriginal=False):
    """
    This method creates a new compressed version of the inputfile compressed with gzip
    Args:
        filename (str): This is the file which is being compressed
        outputfilename (str): This is the (optional) name of the outputfile object which is created
        removeOriginal (bool): Should the original be destroyed after the compressed version is made?
    """
    import gzip
    if not outputfilename:
        outputfilename = filename + '.gz'
    f_in = open(filename, 'rb')
    f_out = gzip.open(outputfilename, 'wb')
    f_out.writelines(f_in)
    f_out.close()
    f_in.close()
    if removeOriginal:
        os.system('rm -f %s' % filename)




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################

from GangaCore.GPIDev.Adapters.IPostProcessor import PostProcessException
from GangaCore.GPIDev.Adapters.IChecker import IFileChecker
from GangaCore.GPIDev.Schema import FileItem, SimpleItem
from GangaCore.Utility.logging import getLogger
import subprocess
import copy
import os
import re


logger = getLogger()


def SortedValues(adict):
    items = sorted(adict.items())
    return [value for key, value in items]


def GetKeyNames(f, dir=""):
    import ROOT
    f.cd(dir)
    return [key.GetName() for key in ROOT.gDirectory.GetListOfKeys()]


def GetTreeObjects(f, dir=""):
    import ROOT
    tree_dict = {}
    for tdir in GetKeyNames(f, dir):
        if tdir == "":
            continue
        absdir = os.path.join(dir, tdir)
        if isinstance(f.Get(tdir), ROOT.TDirectory):
            for absdir, tree in GetTreeObjects(f, absdir).items():
                tree_dict[absdir] = tree
        if isinstance(f.Get(absdir), ROOT.TTree):
            tree_dict[absdir] = f.Get(absdir)
    return tree_dict


class RootFileChecker(IFileChecker):

    """
    Checks ROOT files to see if they are zombies.
    For master job, also checks to see if merging performed correctly.
    self.files are the files you would like to check.
    self.fileMustExist toggles whether to fail the job if the specified file doesn't exist (default is True).
    """
    _schema = IFileChecker._schema.inherit_copy()
    _schema.datadict['checkMerge'] = SimpleItem(
        defvalue=True, doc='Toggle whether to check the merging proceedure')
    _category = 'postprocessor'
    _name = 'RootFileChecker'
    _exportmethods = ['check']

    def checkBranches(self, mastertrees, subtrees):
        import ROOT
        for masterpath, mastertree in mastertrees.items():
            for subpath, subtree in subtrees.items():
                if (subpath == masterpath):
                    subbranches = [branch.GetName()
                                   for branch in subtree.GetListOfBranches()]
                    masterbranches = [branch.GetName()
                                      for branch in mastertree.GetListOfBranches()]
                    if (subbranches != masterbranches):
                        return self.failure
        return self.success

    def addEntries(self, mastertrees, subtrees, entries_dict):
        import ROOT
        for masterpath, mastertree in mastertrees.items():
            for subpath, subtree in subtrees.items():
                if (subpath == masterpath):
                    if (subpath in entries_dict):
                        entries_dict[subpath] += subtree.GetEntries()
                    else:
                        entries_dict[subpath] = subtree.GetEntries()
        return entries_dict

    def checkMergeable(self, f):
        import ROOT
        tf = ROOT.TFile.Open(f)
        if tf.IsZombie():
            logger.info('ROOT file %s is a zombie, failing job', f)
            tf.Close()
            return self.failure
        if not len(GetKeyNames(tf)):
            logger.info('ROOT file %s has no keys, failing job', f)
            tf.Close()
            return self.failure
        tf.Close()
        if (os.path.getsize(f) < 330):
            logger.info('ROOT file %s has no size, failing job', f)
            return self.failure
        return self.success

    def check(self, job):
        """
        Check that ROOT files are not zombies and were closed properly, also (for master job only) checks that the merging performed correctly.
        """
        import ROOT
        self.result = True
        filepaths = self.findFiles(job)
        if self.result is False:
            return self.failure
        if not len(filepaths):
            raise PostProcessException(
                'None of the files to check exist, RootFileChecker will do nothing!')
        for f in filepaths:
            if f.find('.root') < 0:
                raise PostProcessException('The file "%s" is not a ROOT file, RootFileChecker will do nothing!' % os.path.basename(f))
            if not self.checkMergeable(f):
                return self.failure
            if (len(job.subjobs) and self.checkMerge):
                haddoutput = f + '.hadd_output'
                if not os.path.exists(haddoutput):
                    logger.warning('Hadd output file %s does not exist, cannot perform check on merging.', haddoutput)
                    return self.success

                for failString in ['Could not find branch', 'One of the export branches', 'Skipped file']:
                    grepoutput = subprocess.getoutput('grep "%s" %s' % (failString, haddoutput))
                    if len(grepoutput):
                        logger.info('There was a problem with hadd, the string "%s" was found. Will fail job', failString)
                        return self.failure

                tf = ROOT.TFile.Open(f)
                mastertrees = GetTreeObjects(tf)
                entries_dict = {}
                for sj in job.subjobs:
                    if (sj.status == 'completed'):
                        for subfile in self.findFiles(sj):
                            if (os.path.basename(subfile) == os.path.basename(f)):
                                subtf = ROOT.TFile.Open(subfile)
                                subtrees = GetTreeObjects(subtf)

                                substructure = sorted(subtrees.keys())
                                masterstructure = sorted(mastertrees.keys())
                                if (substructure != masterstructure):
                                    logger.info('File structure of subjob %s is not the same as master job, failing job', sj.fqid)
                                    return self.failure

                                if not self.checkBranches(mastertrees, subtrees):
                                    logger.info('The tree structure of subjob %s is not the same as merged tree, failing job', sj.fqid)
                                    return self.failure
                                entries_dict = self.addEntries(
                                    mastertrees, subtrees, entries_dict)
                                subtf.Close()

                master_entries_dict = dict(
                    (n, mastertrees[n].GetEntries()) for n in set(mastertrees))
                if (SortedValues(entries_dict) != SortedValues(master_entries_dict)):
                    logger.info(
                        'Sum of subjob tree entries is not the same as merged tree entries for file %s, failing job (check hadd output)', os.path.basename(f))
                    return self.failure
                tf.Close()
        return self.result






from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from fnmatch import fnmatch
from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile
from GangaCore.Core.exceptions import GangaFileError
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Base.Proxy import isType, GPIProxyObjectFactory
from GangaCore.Utility.Config import getConfig
import re
import copy
import glob
import os
import pickle
import stat
import logging
import GangaCore.Utility.Config

logger = getLogger()
regex = re.compile(r'[*?\[\]]')


class GoogleFile(IGangaFile):

    """
    The GoogleFile outputfile type allows for files to be directly uploaded, downloaded, removed and restored from the GoogleDrive service.
    It can be used as part of a job to output data directly to GoogleDrive, or standalone through the Ganga interface.

    example job: j=Job(application=Executable(exe=File('/home/hep/hs4011/Tests/testjob.sh'), args=[]),outputfiles=[GoogleFile('TestJob.txt')])

                 j.submit()

                 ### This job will automatically upload the outputfile 'TestJob.txt' to GoogleDrive.

    example of standalone submission:

                 g=GoogleFile('TestFile.txt')

                 g.localDir = '~/TestDirectory'        ### The file's location must be specified for standalone submission

                 g.put()                               ### The put() method uploads the file to GoogleDrive directly

    The GoogleFile outputfile is also compatible with the Dirac backend, making outputfiles from Dirac-run jobs upload directly to GoogleDrive.
    """

    _schema = Schema(Version(1, 1),
                     {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                      'localDir': SimpleItem(defvalue="", copyable=1,
                                             doc='local dir where the file is stored, used from get and put methods'),
                      'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1,
                                                sequence=1, copyable=0, doc="collected files from the wildcard namePattern"),
                      'failureReason': SimpleItem(defvalue="", copyable=1,
                                                  doc='reason for the upload failure'),
                      'compressed': SimpleItem(defvalue=False, typelist=[bool], protected=0,
                                               doc='wheather the output file should be compressed before sending somewhere'),
                      'downloadURL': SimpleItem(defvalue="", copyable=1, protected=1,
                                                doc='download URL assigned to the file upon upload to GoogleDrive'),
                      'id': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                       doc='GoogleFile ID assigned to file  on upload to GoogleDrive'),
                      'name': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                                   doc='GoogleFile name of the uploaded file'),
                      'GangaFolderId': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                                  doc='GoogleDrive Ganga folder  ID')
                      })
    _category = 'gangafiles'
    _name = 'GoogleFile'
    _exportmethods = ["get", "put", "remove", "restore", "deleteCredentials", "accessURL"]

    def __init__(self, namePattern=''):
        super(GoogleFile, self).__init__()
        self.namePattern = namePattern
        self.__initialized = False

        self.cred_path = os.path.join(getConfig('Configuration')[
                                      'gangadir'], 'googlecreddata.pkl')

    def __initializeCred(self):
        while os.path.isfile(self.cred_path) == False:
            from google.auth.transport.requests import Request
            from google_auth_oauthlib.flow import InstalledAppFlow

            SCOPES = ['https://www.googleapis.com/auth/drive.file']

            creds = None

            account_details = {
                "installed": {
                    "project_id": "ganga-file-uploader",
                    "token_uri": "https://oauth2.googleapis.com/token",
                    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                    "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                    "client_secret": "-hvpFfe29n5jhUxAXZFlqxxw",
                    "client_id": "893863581947-l7cqdtsa6q9dn8d3cb1lplqsj9odgqia.apps.googleusercontent.com",
                    "redirect_uris": ["urn:ietf:wg:oauth:2.0:oob", "http://localhost"]
                }
            }

            # If there are no (valid) credentials available, let the user log in.
            if not os.path.exists(self.cred_path):
                if not creds or not creds.valid:
                    if creds and creds.expired and creds.refresh_token:
                        creds.refresh(Request())
                        logger.info(
                            'Enter you accound details in the browser window prompted')
                    else:
                            flow = InstalledAppFlow.from_client_config(
                                account_details, 
                                SCOPES
                            )                        
                            creds = flow.run_local_server(port=0)

                    # Save the credentials for the next run
                    with open(self.cred_path, 'wb') as token:
                        pickle.dump(creds, token)

                    os.chmod(self.cred_path, stat.S_IWUSR | stat.S_IRUSR)
                    logger.info('Your GoogleDrive credentials have been stored in the file %s and are only readable by you. '
                                'The file will give permission to modify files in your GoogleDrive. '
                                'Permission can be revoked by going to "Manage Apps" in your GoogleDrive '
                                'or by deleting the credentials through the deleteCredentials GoogleFile method.' % self.cred_path)

        self.__initialized = True
        self._check_Ganga_folder()

    def _attribute_filter__set__(self, n, v):
        if n == 'localDir':
            return os.path.expanduser(os.path.expandvars(v))
        return v

    def setLocation(self):
        """
        Sets the location of output files that were uploaded from the WN
        """
        raise NotImplementedError

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        raise NotImplementedError

    def _on_attribute__set__(self, obj_type, attrib_name):
        r = copy.deepcopy(self)
        from GangaCore.GPIDev.Lib.Job import Job
        if isinstance(obj_type, Job) and attrib_name == 'outputfiles':
            r.localDir = None
            r.failureReason = ''
        return r

    def deleteCredentials(self):
        """
        Deletes the user's GoogleDrive credentials

            example use: GoogleFile().deleteCredentials()
        """
        if os.path.isfile(self.cred_path) == True:
            os.remove(self.cred_path)
            logger.info('GoogleDrive credentials deleted')
            return None
        else:
            logger.info('There are no credentials to delete')

    def internalCopyTo(self, targetPath):
        """
        Retrieves files uploaded to GoogleDrive
        Args:
            targetPath (str): Target path where the file is copied to
        """
        import io
        from googleapiclient.http import MediaIoBaseDownload

        dir_path = targetPath
        service = self._setup_service()

        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

        # Checks for wildcards and loops through get procedure for each result,
        # saving file to assigned directory
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                if f.id:
                    completeName = os.path.join(dir_path, f.name)
                    self.download_file_from_drive(service, f.id, completeName)

                else:
                    logger.info("Download unsuccessful, file \'%s\' may not exist on GoogleDrive" % f.name)

        # Non-wildcard get request procedure
        else:
            if self.id:
                completeName = os.path.join(dir_path, self.name)
                self.download_file_from_drive(service, self.id, completeName)

            else:
                # print 'An error occurred: %s' % resp
                logger.info(
                    "Download unsuccessful, the file may not exist on GoogleDrive")
                return None

    def getWNScriptDownloadCommand(self, indent):
        """
        Gets the command used to download already uploaded file
        """
        raise NotImplementedError

    def processWildcardMatches(self):
        raise NotImplementedError

    def __repr__(self):
        """
        Get the representation of the file
        """
        return "GoogleFile(namePattern='%s', downloadURL='%s')" % (self.namePattern, self.downloadURL)

    def accessURL(self):
        if self.subfiles:
            URLs = []
            for f in self.subfiles:
                URLs.append(f.downloadURL)
        elif self.id:
            URLs = self.downloadURL
        return URLs

    def download_file_from_drive(self, service, fileid, filepath, filename=None):
        import io
        from googleapiclient.http import MediaIoBaseDownload

        # if file name is not known, we first get the file's name
        if filename == None:
            name_request = service.files().get(fileId=fileid).execute()
            fname = name_request['name']
            filename = os.path.join(filepath, fname)

        request = service.files().get_media(fileId=fileid)
        fh = io.FileIO(filename, 'wb')
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            logger.info(f"Downloading file: {self.name} {int(status.progress()*100)}")
        logger.info("Download successful")

    def get(self):
        """
        Method to get the Local file from files uploaded to GoogleDrive by ganga
        """
        service = self._setup_service()

        # Sets the target directory
        dir_path = self.localDir
        if self.localDir == '':
            dir_path = os.getcwd()

        if self._getParent() is not None:
            dir_path = self.getJobObject().getOutputWorkspace().getPath()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            import fnmatch

            files_found = False
            file_regex = fnmatch.translate(self.namePattern)

            search_result = service.files().list(
                q=f"'{self.GangaFolderId}' in parents",
                spaces='drive'
            ).execute() 

            if search_result['files']:
                for _file in search_result['files']:
                    if re.match(file_regex, _file['name']) is not None:
                        files_found = True
                        self.download_file_from_drive(
                            service, _file['id'], dir_path, None
                        )
                if files_found is False:
                    raise GangaFileError(f"No files with pattern: {self.namePattern} were found in Ganga Folder of your Google Drive. Ganga can only see files that was uploded by Ganga itself.")
            else:
                raise GangaFileError(f"Ganga Folder of your Google Drive is empty/non-existent")
        else:
            search_result = service.files().list(
                q=f"name = '{self.namePattern}' and parents in '{self.GangaFolderId}'",
                spaces='drive'
            ).execute() 

            if search_result['files']:
                for _file in search_result['files']:
                    self.download_file_from_drive(
                        service, _file['id'], dir_path, self.namePattern
                    )
            else:
                raise GangaFileError(f"File: {self.namePattern} not found in Ganga Folder of your Google Drive. Ganga can only see files that was uploded by Ganga itself.")

    def put(self):
        """
        Postprocesses (upload) output file to the desired destination from the client
        """
        import hashlib
        from googleapiclient.http import MediaFileUpload

        service = self._setup_service()

        # Sets the target directory
        dir_path = self.localDir
        if self.localDir == '':
            dir_path = os.getcwd()

        if self._getParent() is not None:
            dir_path = self.getJobObject().getOutputWorkspace().getPath()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for wildfile in glob.glob(os.path.join(dir_path, self.namePattern)):
                FILENAME = wildfile
                filename = os.path.basename(wildfile)
                # checking if the to be uploaded exists
                if not os.path.isfile(FILENAME):
                    raise GangaFileError(f"File: {FILENAME} not found.")

                file_metadata = {
                    'name': filename,
                    'description': 'A test document',
                    'mimeType': 'text/plain',
                    'parents': [self.GangaFolderId]
                }
                media = MediaFileUpload(
                    FILENAME,
                    mimetype='application/vnd.google-apps.document'
                )
                file = service.files().create(
                    fields='id',
                    media_body=media,
                    body=file_metadata
                ).execute()

                # Checking the hash of inserted data
                with open(FILENAME, 'rb') as thefile:
                    file_results = service.files().list(
                        q=f"name='{filename}'",
                        fields="nextPageToken, files(id, name, md5Checksum)"
                    ).execute()

                    for _file in file_results.get('files', []):
                        # Found the file's information that was just uploaded
                        if _file['id'] == file['id']:
                            if _file['md5Checksum'] == hashlib.md5(thefile.read()).hexdigest():
                                logger.info("File \'%s\' uploaded succesfully" %
                                            filename)
                            else:
                                raise GangaFileError("Upload of \'%s\' unsuccessful" % filename)

                # Assign new schema components to each file and append to job
                # subfiles
                g = GoogleFile(filename)
                g.downloadURL = f"https://drive.google.com/file/d/{file['id']}"
                g.id = file['id']
                g.name = file_metadata['name']
                self.subfiles.append(GPIProxyObjectFactory(g))

        # For non-wildcard upload
        else:
            FILENAME = os.path.join(dir_path, self.namePattern)
            if not os.path.isfile(FILENAME):
                raise GangaFileError(f"File: {FILENAME} not found.")

            file_metadata = {
                'name': self.namePattern,
                'description': 'A test document',
                'mimeType': 'text/plain',
                'parents': [self.GangaFolderId]
            }
            media = MediaFileUpload(
                FILENAME,
                mimetype='application/vnd.google-apps.document'
            )
            file = service.files().create(
                fields='id',
                media_body=media,
                body=file_metadata
            ).execute()

            # Checking the hash of inserted data
            with open(FILENAME, 'rb') as thefile:
                file_results = service.files().list(
                    q=f"name='{self.namePattern}'",
                    fields="nextPageToken, files(id, name, md5Checksum)"
                ).execute()

                for _file in file_results.get('files', []):
                    # Found the file's information that was just uploaded
                    if _file['id'] == file['id']:
                        if _file['md5Checksum'] == hashlib.md5(thefile.read()).hexdigest():
                            logger.info("File \'%s\' uploaded succesfully" %
                                        self.namePattern)
                        else:
                            raise GangaFileError("Upload of \'%s\' unsuccessful" % self.namePattern)

            # Assign values to new schema components
            self.downloadURL = f"https://drive.google.com/file/d/{file['id']}"
            self.id = file['id']
            self.name = file_metadata['name']
            return
            
        return GPIProxyObjectFactory(self.subfiles[:])

    def remove(self, permanent=False):
        """
        Move a file to the trash or permanently delete the file

            example use: GoogleFile().remove()

            or:          j = Job([...], outputfiles=GoogleFile()) --> j.submit --> j.outputfiles[0].remove()

        Remove multiple files by using

                         for i in j.outputfiles:
                             i.remove()

        The file can also be permanently deleted by using

                         GoogleFile().remove(True)

        However, this will make the file unrestorable
        """
        from googleapiclient.errors import HttpError 
        service = self._setup_service()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                if permanent == True:
                    try:
                        service.files().delete(fileId=f.id).execute()
                        f.downloadURL = ''
                        logger.info(
                            'File \'%s\' permanently deleted from GoogleDrive' % f.name)
                    except HttpError as error:
                        # print 'An error occurred: %s' % error
                        logger.info(
                            'File \'%s\' deletion failed, or file already deleted' % f.name)
                else:
                    try:
                        # updating the file metadata to trash it
                        service.files().update(
                            fileId=f.id,
                            body={"trashed": True}
                        ).execute()
                        logger.info(
                            'File \'%s\' removed from GoogleDrive, added to trash' % f.name)
                    except HttpError as error:
                        # print 'An error occurred: %s' % error
                        logger.info(
                            'File \'%s\' removal failed, or file already removed' % f.name)

        # Non-wildcard request
        else:
            if permanent == True:
                try:
                    service.files().delete(fileId=self.id).execute()
                    self.downloadURL = ''
                    logger.info('File permanently deleted from GoogleDrive')
                except HttpError as error:
                    logger.info(
                        'File deletion failed, or file already deleted')
            else:
                try:
                    # updating the file metadata to trash it
                    service.files().update(
                        fileId=self.id,
                        body={"trashed": True}
                    ).execute()
                    logger.info('File removed from GoogleDrive, added to trash')
                except HttpError as error:
                    logger.info('File removal failed, or file already removed')
                return None

    def restore(self):
        """
        Restore a file from the trash. This method will not work on permanently deleted files

            example use: GoogleFile().restore()
        """
        from googleapiclient.errors import HttpError 

        service = self._setup_service()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                try:
                    service.files().update(
                        fileId=f.id,
                        body={"trashed": False}
                    ).execute()                 
                    logger.info(
                        'File \'%s\' restored to GoogleDrive' % f.name)
                except HttpError as error:
                    # print 'An error occurred: %s' % error
                    logger.info(
                        'File \'%s\' restore failed, or file does not exist on GoogleDrive' % f.name)
                return None
        # Non-wildcard request
        else:
            try:
                service.files().update(
                    fileId=self.id,
                    body={"trashed": False}
                ).execute()
                logger.info(
                        'File \'%s\' restored to GoogleDrive' % self.name)
            except HttpError as error:
                # print 'An error occurred: %s' % error
                logger.info(
                        'File \'%s\' restore failed, or file does not exist on GoogleDrive' % self.name)
            return None

    def _check_Ganga_folder(self):
        """
        Creates a Ganga folder on GoogleDrive if one is not already present
        """
        service = self._setup_service()

        # grabing all the folders in root folder of gdrive 
        results = service.files().list(
            pageSize=10, fields="nextPageToken, files(id, name)"
        ).execute()
        items = results.get('files', [])

        for _file in items:
            if _file['name'] == 'Ganga':
                self.GangaFolderId = _file['id']
                
        if not self.GangaFolderId:
            body = {
                'name': 'Ganga',
                'description': 'A test folder',
                'mimeType': 'application/vnd.google-apps.folder'
            }
            file = service.files().create(body=body).execute()
            self.GangaFolderId = file.get('id')

    def _setup_service(self):
        """
        Sets up the GoogleDrive service for other methods
        """
        from googleapiclient.discovery import build
        if self.__initialized == False:
            self.__initializeCred()
        with open(self.cred_path, "rb") as nput:
            credentials = pickle.load(nput)
        service = build('drive', 'v3', credentials=credentials, cache_discovery=False)
        return service

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        logger.info('injecting')

    def _readonly(self):
        return False

    def _list_get__match__(self, to_match):
        if isinstance(to_match, str):
            return fnmatch(self.namePattern, to_match)
        if isinstance(to_match, type):
            # note stripProxy wont work on class types that aren't instances
            return isinstance(self, to_match._impl)
        return to_match == self

GangaCore.Utility.Config.config_scope['GoogleFile'] = GoogleFile



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ExeSplitter.py,v 1.1 2008-07-17 16:40:59 moscicki Exp $
##########################################################################

from GangaCore.GPIDev.Schema import Schema, Version, ComponentItem
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter


class ExeSplitter(ISplitter):

    """ Split executable applications (OBSOLETE).

    This splitter allows the creation of subjobs where each subjob has a different Executable application.
    This splitter is OBSOLETED use GenericSplitter or ArgSplitter instead.
    """
    _name = "ExeSplitter"
    _schema = Schema(Version(1, 0), {
        'apps': ComponentItem('applications', defvalue=[], sequence=1, doc='a list of Executable app objects')
    })

    def split(self, job):
        subjobs = []
        for a in self.apps:
            # for each subjob make a full copy of the master job
            j = self.createSubjob(job)
            j.application = a
            if not a.exe:
                j.application.exe = job.application.exe
            subjobs.append(j)
        return subjobs


#
#
# $Log: not supported by cvs2svn $
# Revision 1.3  2007/07/10 13:08:32  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.2  2006/08/24 16:52:10  moscicki
# splitter.createJob()
#
# Revision 1.1  2006/06/21 11:27:29  moscicki
# splitters moved to the new location
#
#
#



################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 10/02/2014
################################################################################
"""@package RecoPlusVFT
This module is designed to run any ND280 executable accessible in the $PATH environment variable.
"""

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

import os, shutil, re, time
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class RecoPlusVFT(IPrepareApp):
    """
    RecoPlusVFT application running any ND280 executables.

    The required input for this module are:
        app.reco_exe = 'RunOARecon.exe'
        app.vft_exe = 'OAReconValidTree.exe'
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'

    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to analyze only 10 events:
        app.reco_args = ['-n',10]
    """
    _schema = Schema(Version(1,1), {
        'cmtsetup' : SimpleItem(defvalue=None,doc='Setup script in bash to set up cmt and the cmt package of the executable.', typelist=['str','type(None)']),
        'reco_exe' : SimpleItem(defvalue=None,typelist=['str', 'type(None)'],comparable=1,doc='A path (string) or a File object specifying the reconstruction executable. For example oaRecon.exe.'), 
        'reco_args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'vft_only' : SimpleItem(defvalue=True,doc='Run only VFT on the input files, i.e. don\'t call a reconstruction software.', typelist=['bool']),
        'vft_exe' : SimpleItem(defvalue=None,typelist=['str', 'type(None)'],comparable=1,doc='A path (string) or a File object specifying the validation flat tree executable. For example OAReconValidTree.exe.'), 
        'vft_args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'filenamesubstr' : SimpleItem(defvalue=None,doc='This string will be substituted by recnewstr variable in the reco output filename and vftnewstr variable in the VFT output filename.', typelist=['str','type(None)']),
        'reconewstr' : SimpleItem(defvalue='newreco',doc='This string will substitute filenamesubstr in the input filename to create the reco output filename.', typelist=['str']),
        'vftnewstr' : SimpleItem(defvalue='validtree',doc='This string will substitute filenamesubstr in the input filename to create the VFT output filename.', typelist=['str']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'RecoPlusVFT'
    _scriptname = None
    _exportmethods = ['prepare']
    _GUIPrefs = [ { 'attribute' : 'reco_exe', 'widget' : 'File' },
                  { 'attribute' : 'reco_args', 'widget' : 'String_List' },
                  { 'attribute' : 'vft_exe', 'widget' : 'File' },
                  { 'attribute' : 'vft_args', 'widget' : 'String_List' },
                  { 'attribute' : 'filenamesubstr', 'widget' : 'String' },
                  { 'attribute' : 'reconewstr', 'widget' : 'String' },
                  { 'attribute' : 'vftnewstr', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'reco_exe', 'widget' : 'File' },
                          { 'attribute' : 'reco_args', 'widget' : 'String_List' },
                          { 'attribute' : 'vft_exe', 'widget' : 'File' },
                          { 'attribute' : 'vft_args', 'widget' : 'String_List' },
                          { 'attribute' : 'filenamesubstr', 'widget' : 'String' },
                          { 'attribute' : 'reconewstr', 'widget' : 'String' },
                          { 'attribute' : 'vftnewstr', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    def __init__(self):
        super(RecoPlusVFT,self).__init__()


    def configure(self,masterappconfig):
        if self.cmtsetup is None:
          raise ApplicationConfigurationError('No cmt setup script given.')

        # __________ Reco first ____________
        reco_args = convertIntToStringArgs(self.reco_args)

        job = self.getJobObject()

        # Need to handle the possibility of multiple output files !
        # setup the output file
        for arg in reco_args:
          if arg == '-o':
            raise ApplicationConfigurationError('Option "-o" given in reco_args. You must use the filenamesubstr and reconewstr variables instead to define an output.')

        # So get the list of filenames get_dataset_filenames() and create a file containing the list of files and put it in the sandbox
        if job.inputdata is None:
          raise ApplicationConfigurationError('The inputdata variable is not defined.')
        fileList = job.inputdata.get_dataset_filenames()
        if len(fileList) < 1:
          raise ApplicationConfigurationError('No input data file given.')

        firstFile = fileList[0].split('/')[-1]
        # Define the output
        reco_args.append('-o')
        if self.filenamesubstr is None:
          reco_outputfile = 'recoOutput.root'
        else:
          reco_outputfile = firstFile.replace(self.filenamesubstr, self.reconewstr)

        reco_args.append(reco_outputfile)
        
        # Just to define the output before the potentially long list of input files
        reco_args.extend(fileList)

        # __________ Now VFT ____________
        vft_args = convertIntToStringArgs(self.vft_args)

        job = self.getJobObject()

        # Need to handle the possibility of multiple output files !
        # setup the output file
        for arg in vft_args:
          if arg == '-o':
            raise ApplicationConfigurationError('Option "-o" given in vft_args. You must use the filenamesubstr and reconewstr variables instead to define an output.')

        # Define the output
        vft_args.append('-o')
        if self.filenamesubstr is None:
          vft_outputfile = 'vftOutput.root'
        else:
          vft_outputfile = firstFile.replace(self.filenamesubstr, self.vftnewstr)
        vft_args.append(vft_outputfile)

        # Use the reco output as an input for the VFT processing
        # or use the input file list if running in VFT only mode.
        if self.vft_only:
            vft_args.extend(fileList)
        else:
            vft_args.append(reco_outputfile)

        reco_argsStr = ' '.join(reco_args)
        vft_argsStr = ' '.join(vft_args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        script += 'source '+self.cmtsetup+'\n'
        if not self.vft_only:
            script += 'echo "RECO EXE ..."\n'
            script += self.reco_exe+' '+reco_argsStr+' 2>&1\n'
            script += 'echo "... RECO EXE"\n'
        script += 'echo "VFT EXE ..."\n'
        script += self.vft_exe+' '+vft_argsStr+' 2>&1\n'
        script += 'echo "... VFT EXE"\n'

        from GangaCore.GPIDev.Lib.File import FileBuffer

        scriptname = 'RecoPlusVFT.sh'
        job.getInputWorkspace().writefile(FileBuffer(scriptname,script),executable=1)

        self._scriptname = job.inputdir+scriptname

        return (None,None)



config = getConfig('defaults_RecoPlusVFT') #_Properties
config.options['reco_exe'].type = type(None)
config.options['vft_exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        return StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('RecoPlusVFT','LSF', RTHandler)
allHandlers.add('RecoPlusVFT','Local', RTHandler)
allHandlers.add('RecoPlusVFT','PBS', RTHandler)
allHandlers.add('RecoPlusVFT','SGE', RTHandler)
allHandlers.add('RecoPlusVFT','Slurm', RTHandler)
allHandlers.add('RecoPlusVFT','Condor', RTHandler)
allHandlers.add('RecoPlusVFT','LCG', LCGRTHandler)
allHandlers.add('RecoPlusVFT','gLite', gLiteRTHandler)
allHandlers.add('RecoPlusVFT','TestSubmitter', RTHandler)
allHandlers.add('RecoPlusVFT','Interactive', RTHandler)
allHandlers.add('RecoPlusVFT','Batch', RTHandler)
allHandlers.add('RecoPlusVFT','Cronus', RTHandler)
allHandlers.add('RecoPlusVFT','Remote', LCGRTHandler)
allHandlers.add('RecoPlusVFT','CREAM', LCGRTHandler)




##########################################################################
# Ganga Project. https://github.com/ganga-devs/ganga
#
##########################################################################
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.Utility.Config import getConfig
from GangaCore.GPIDev.Adapters.IVirtualization import IVirtualization


class Docker(IVirtualization):

    """
    The job will be run inside a container using Docker or UDocker as the virtualization method. Docker
    is tried first and if not installed or permission do not allow it, UDocker is installed and used.

    j=Job()
    j.virtualization = Docker("fedora:latest")   

    The mode of the UDocker running can be modified. The P1 mode is working almost everywhere but might
    not give the best performance. See https://github.com/indigo-dc/udocker for more details about
    Udocker.

    If the image is a private image, the username and password of the deploy token can be given like

    j.virtualization.tokenuser = 'gitlab+deploy-token-123'
    j.virtualization.tokenpassword = 'gftrh84dgel-245^ghHH'

    Note that images stored in a docker repository hosted by Github at present doesn't work with uDocker 
    as uDocker is not updated to the latest version of the API.

    Directories can be mounted from the host to the container using key-value pairs to the mounts option.

    j.virtualization.mounts = {'/cvmfs':'/cvmfs'}
    """
    _name = 'Docker'
    _schema = IVirtualization._schema.inherit_copy()
    _schema.datadict['mode'] = SimpleItem(defvalue="P1", doc='Mode of container execution')

    def __init__(self, image, mode):
        super().__init__(image)
        self.mode = mode

    def modify_script(self, script, sandbox=False):
        """Overides parent's modify_script function
            Arguments other than self:
               script - Script that need to be modified

            Return value: modified script"""

        extra = 'virtualization_image = ' + repr(self.image) + '\n'
        extra = extra + 'virtualization_user = ' + repr(self.tokenuser) + '\n'
        extra = extra + 'virtualization_password = ' + repr(self.tokenpassword) + '\n'
        extra = extra + 'virtualization_mounts = ' + repr(self.mounts) + '\n'
        extra = extra + 'virtualization_options = ' + repr(self.options) + '\n'

        if sandbox:
            extra = extra + 'virtualization_udockerlocation = ' + repr(getcwd())
            extra = extra + 'runenv[\'UDOCKER_DIR\']=' + repr(path.join(getcwd(),'.udocker'))
        else:
            extra = extra + 'virtualization_udockerlocation = ' + \
            repr(getConfig('Configuration')['UDockerlocation'])  + '\n'
        
        extra = extra + """

from Virtualization import checkDocker, checkUDocker, checkSingularity, installUdocker
options = []
                
if execmd[0].startswith('./'): 
    execmd[0] = "/work_dir/"+execmd[0]
if (checkDocker()):
    print("Using Docker")
    if virtualization_user:
        buildcommand = ['docker', 'login', '--username='+virtualization_user, '--password='+virtualization_password]
        rc = subprocess.call(buildcommand, env=runenv, shell=False)
    for k,v in virtualization_mounts.items():
        if os.path.isdir(k):
            options = options + ['-v' , k + ':' + v]
        else:
            print('Requested directory %s is not available and no bind will be made to container' % k)
    options = options + virtualization_options
    execmd = ['docker', 'run', '--rm', '-v', workdir+":"+"/work_dir"] + options + [virtualization_image] + execmd        
else:
    print("Docker not available or no permission to run docker demon, will attempt UDocker.")
    location = os.path.expanduser(virtualization_udockerlocation)
    binary = os.path.join(location,'udocker')
    if not (checkUDocker(location)):
        try:
            installUdocker(location)
        except OSError as x:
            failurereport(statusfile, 'PROBLEM WITH UDOCKER: %s' % str(x))
    runenv["PROOT_NO_SECCOMP"]="1"
    runenv['UDOCKER_DIR']=os.path.join(location,'.udocker')
    if virtualization_user:
        buildcommand = [binary, 'login', '--username='+virtualization_user, '--password='+virtualization_password]
        rc = subprocess.call(buildcommand, env=runenv, shell=False)
    for k,v in virtualization_mounts.items():
        if os.path.isdir(k):
            options = options + ['--volume='+ k + ':' + v]
        else:
            print('Requested directory %s is not available and no bind will be made to container' % k)
    options = options + virtualization_options
    execmd = [binary, '--quiet', 'run', '--rm', '--nosysdirs', '--volume', workdir+":"+"/work_dir"] + options + [virtualization_image] + execmd

"""
        script = script.replace('###VIRTUALIZATION###',extra)
        return script



#----------------------------------------------------------------------
#  eliza.py
#
#  a cheezy little Eliza knock-off by Joe Strout <joe@strout.net>
#  with some updates by Jeff Epler <jepler@inetnebr.com>
#  hacked into a module and updated by Jez Higgins <jez@jezuk.co.uk>
#  last revised: 28 February 2005
#----------------------------------------------------------------------

import string
import re
import random


class eliza:

    def __init__(self):
        self.keys = [re.compile(x[0], re.IGNORECASE) for x in gPats]
        self.values = [x[1] for x in gPats]

    #----------------------------------------------------------------------
    # translate: take a string, replace any words found in dict.keys()
    #  with the corresponding dict.values()
    #----------------------------------------------------------------------
    def translate(self, str, dict):
        words = string.split(string.lower(str))
        keys = list(dict.keys())
        for i in range(0, len(words)):
            if words[i] in keys:
                words[i] = dict[words[i]]
        return string.join(words)

    #----------------------------------------------------------------------
    #  respond: take a string, a set of regexps, and a corresponding
    #    set of response lists; find a match, and return a randomly
    #    chosen response from the corresponding list.
    #----------------------------------------------------------------------
    def respond(self, str):
        # find a match among keys
        for i in range(0, len(self.keys)):
            match = self.keys[i].match(str)
            if match:
                # found a match ... stuff with corresponding value
                # chosen randomly from among the available options
                resp = random.choice(self.values[i])
                # we've got a response... stuff in reflected text where
                # indicated
                pos = string.find(resp, '%')
                while pos > -1:
                    num = string.atoi(resp[pos + 1:pos + 2])
                    resp = resp[:pos] + \
                        self.translate(match.group(num), gReflections) + \
                        resp[pos + 2:]
                    pos = string.find(resp, '%')
                # fix munged punctuation at the end
                if resp[-2:] == '?.':
                    resp = resp[:-2] + '.'
                if resp[-2:] == '??':
                    resp = resp[:-2] + '?'
                return resp

#----------------------------------------------------------------------
# gReflections, a translation table used to convert things you say
#    into things the computer says back, e.g. "I am" --> "you are"
#----------------------------------------------------------------------
gReflections = {
    "am": "are",
    "was": "were",
    "i": "you",
    "i'd": "you would",
    "i've": "you have",
    "i'll": "you will",
    "my": "your",
    "are": "am",
    "you've": "I have",
    "you'll": "I will",
    "your": "my",
    "yours": "mine",
    "you": "me",
    "me": "you"
}

#----------------------------------------------------------------------
# gPats, the main response table.  Each element of the list is a
#  two-element list; the first is a regexp, and the second is a
#  list of possible responses, with group-macros labelled as
#  %1, %2, etc.
#----------------------------------------------------------------------
gPats = [
    [r'I need (.*)',
     ["Why do you need %1?",
      "Would it really help you to get %1?",
      "Are you sure you need %1?"]],

    [r'Why don\'?t you ([^\?]*)\??',
     ["Do you really think I don't %1?",
      "Perhaps eventually I will %1.",
      "Do you really want me to %1?"]],

    [r'Why can\'?t I ([^\?]*)\??',
        ["Do you think you should be able to %1?",
         "If you could %1, what would you do?",
         "I don't know -- why can't you %1?",
         "Have you really tried?"]],

    [r'I can\'?t (.*)',
        ["How do you know you can't %1?",
         "Perhaps you could %1 if you tried.",
         "What would it take for you to %1?"]],

    [r'I am (.*)',
        ["Did you come to me because you are %1?",
         "How long have you been %1?",
         "How do you feel about being %1?"]],

    [r'I\'?m (.*)',
        ["How does being %1 make you feel?",
         "Do you enjoy being %1?",
         "Why do you tell me you're %1?",
         "Why do you think you're %1?"]],

    [r'Are you ([^\?]*)\??',
        ["Why does it matter whether I am %1?",
         "Would you prefer it if I were not %1?",
         "Perhaps you believe I am %1.",
         "I may be %1 -- what do you think?"]],

    [r'What (.*)',
        ["Why do you ask?",
         "How would an answer to that help you?",
         "What do you think?"]],

    [r'How (.*)',
        ["How do you suppose?",
         "Perhaps you can answer your own question.",
         "What is it you're really asking?"]],

    [r'Because (.*)',
        ["Is that the real reason?",
         "What other reasons come to mind?",
         "Does that reason apply to anything else?",
         "If %1, what else must be true?"]],

    [r'(.*) sorry (.*)',
        ["There are many times when no apology is needed.",
         "What feelings do you have when you apologize?"]],

    [r'Hello(.*)',
        ["Hello... I'm glad you could drop by today.",
         "Hi there... how are you today?",
         "Hello, how are you feeling today?"]],

    [r'I think (.*)',
        ["Do you doubt %1?",
         "Do you really think so?",
         "But you're not sure %1?"]],

    [r'(.*) friend (.*)',
        ["Tell me more about your friends.",
         "When you think of a friend, what comes to mind?",
         "Why don't you tell me about a childhood friend?"]],

    [r'Yes',
        ["You seem quite sure.",
         "OK, but can you elaborate a bit?"]],

    [r'(.*) computer(.*)',
        ["Are you really talking about me?",
         "Does it seem strange to talk to a computer?",
         "How do computers make you feel?",
         "Do you feel threatened by computers?"]],

    [r'Is it (.*)',
        ["Do you think it is %1?",
         "Perhaps it's %1 -- what do you think?",
         "If it were %1, what would you do?",
         "It could well be that %1."]],

    [r'It is (.*)',
        ["You seem very certain.",
         "If I told you that it probably isn't %1, what would you feel?"]],

    [r'Can you ([^\?]*)\??',
        ["What makes you think I can't %1?",
         "If I could %1, then what?",
         "Why do you ask if I can %1?"]],

    [r'Can I ([^\?]*)\??',
        ["Perhaps you don't want to %1.",
         "Do you want to be able to %1?",
         "If you could %1, would you?"]],

    [r'You are (.*)',
        ["Why do you think I am %1?",
         "Does it please you to think that I'm %1?",
         "Perhaps you would like me to be %1.",
         "Perhaps you're really talking about yourself?"]],

    [r'You\'?re (.*)',
        ["Why do you say I am %1?",
         "Why do you think I am %1?",
         "Are we talking about you, or me?"]],

    [r'I don\'?t (.*)',
        ["Don't you really %1?",
         "Why don't you %1?",
         "Do you want to %1?"]],

    [r'I feel (.*)',
        ["Good, tell me more about these feelings.",
         "Do you often feel %1?",
         "When do you usually feel %1?",
         "When you feel %1, what do you do?"]],

    [r'I have (.*)',
        ["Why do you tell me that you've %1?",
         "Have you really %1?",
         "Now that you have %1, what will you do next?"]],

    [r'I would (.*)',
        ["Could you explain why you would %1?",
         "Why would you %1?",
         "Who else knows that you would %1?"]],

    [r'Is there (.*)',
        ["Do you think there is %1?",
         "It's likely that there is %1.",
         "Would you like there to be %1?"]],

    [r'My (.*)',
        ["I see, your %1.",
         "Why do you say that your %1?",
         "When your %1, how do you feel?"]],

    [r'You (.*)',
        ["We should be discussing you, not me.",
         "Why do you say that about me?",
         "Why do you care whether I %1?"]],

    [r'Why (.*)',
        ["Why don't you tell me the reason why %1?",
         "Why do you think %1?"]],

    [r'I want (.*)',
        ["What would it mean to you if you got %1?",
         "Why do you want %1?",
         "What would you do if you got %1?",
         "If you got %1, then what would you do?"]],

    [r'(.*) mother(.*)',
        ["Tell me more about your mother.",
         "What was your relationship with your mother like?",
         "How do you feel about your mother?",
         "How does this relate to your feelings today?",
         "Good family relations are important."]],

    [r'(.*) father(.*)',
        ["Tell me more about your father.",
         "How did your father make you feel?",
         "How do you feel about your father?",
         "Does your relationship with your father relate to your feelings today?",
         "Do you have trouble showing affection with your family?"]],

    [r'(.*) child(.*)',
        ["Did you have close friends as a child?",
         "What is your favorite childhood memory?",
         "Do you remember any dreams or nightmares from childhood?",
         "Did the other children sometimes tease you?",
         "How do you think your childhood experiences relate to your feelings today?"]],

    # OK, lets add a few ganga specific keywords so we can have some fun
    # Needs to be really extened to include som god questions. -- Andrew
    [r'(.*)submit(.*)',
        ["Why do you want to submit?",
         "Does submitting give you pleasure?",
         "Did you ever try an alternative to submitting?"]],

    [r'(.*)job(.*)',
        ["Is your job important to you?",
         "Why do you need to do this job?",
         "Does losing this job worry you?"]],

    [r'(.*)Gaudi(.*)',
        ["Do you like Spanish architecture?",
         "Are you afraid, like Gaudi, of not completing your work?",
         "Did your interest in Gaudi start as a child?"]],

    [r'(.*)Ganga(.*)',
        ["Are you a reborn Indian?",
         "Is your interest in the Ganges river related to Indian cuisine?"]],

    [r'(.*)backend(.*)',
        ["Why not use a frontend?",
         "Has it anything to do with your father?"]],

    [r'(.*)Dial(.*)',
        ["Why don't you give me your number?",
         "Am I your interactive help or your date?",
         "Why not try to call the helpdesk?"]],

    [r'(.*)Dirac(.*)',
        ["Why do you want to talk about a dead physicist?",
         "Do you feel preasure, because you are expected to win a Nobel prize?",
         "Are you interested in science?"]],

    [r'help',
        ["Do you think this help is not useful?",
         "Is this a cry for help?",
         "Why don't you try to be more verbose?"]],

    [r'(.*)\?',
        ["Why do you ask that?",
         "Please consider whether you can answer your own question.",
         "Perhaps the answer lies within yourself?",
         "Why don't you tell me?"]],

    [r'quit',
        ["Thank you for talking with me.",
         "Good-bye.",
         "Thank you, that will be $150.  Have a good day!"]],

    [r'(.*)',
        ["Please tell me more.",
         "Let's change focus a bit... Tell me about your family.",
         "Can you elaborate on that?",
         "Why do you say that %1?",
         "I see.",
         "Very interesting.",
         "%1.",
         "I see.  And what does that tell you?",
         "How does that make you feel?",
         "How do you feel when you say that?"]]
]

#----------------------------------------------------------------------
#  command_interface
#----------------------------------------------------------------------


def command_interface():
    print("Ganga Interactive Help\n\n")
    print("Ask your questions by typing them in plain English, using normal upper-")
    print('and lower-case letters and punctuation.  Enter "quit" when done.')
    print('=' * 72)
    print("Hello.  What is your problem?\n")
    s = ""
    therapist = eliza()
    while s != "quit":
        try:
            s = input(">")
        except EOFError:
            s = "quit"
            print(s)
        while s[-1] in "!.":
            s = s[:-1]
        print(therapist.respond(s))


if __name__ == "__main__":
    command_interface()


#
#
# $Log: not supported by cvs2svn $
# Revision 1.3  2007/01/25 15:52:39  moscicki
# mergefrom_Ganga-4-2-2-bugfix-branch_25Jan07 (GangaBase-4-14)
#
# Revision 1.2.4.1  2007/01/22 14:20:10  moscicki
# removed deprecated whrandom module
#
# Revision 1.2  2005/08/31 15:21:15  andrew
# More help
#
#
#



"""Utilities for GangaRobot."""

from GangaCore.Utility.Config.Config import getConfig
from GangaCore.Utility import files
import os
import time

ID_FORMAT = '%Y-%m-%d_%H.%M.%S'
TIME_FORMAT = '%Y/%m/%d %H:%M:%S'


def getconfig():
    """Return the [Robot] section of the global configuration."""
    return getConfig('Robot')


def utcid(utctime = None):
    """Return a UTC ID string based on the given UTC time or now if not given.
    
    Keyword arguments:
    utctime -- Optional UTC time string. See utctime().
    
    """
    if utctime:
        return utctime.replace('/', '-').replace(' ', '_').replace(':', '.')
    else:
        return time.strftime(ID_FORMAT,time.gmtime())


def utctime(utcid = None):
    """Return a UTC time string based on the given UTC ID or now if not given.
    
    Keyword arguments:
    utcid -- Optional UTC ID string. See utcid().
    
    """
    if utcid:
        return utcid.replace('-', '/').replace('_', ' ').replace('.', ':')
    else:
        return time.strftime(TIME_FORMAT,time.gmtime())


def jobtreepath(utcid):
    """Return the JobTree path corresponding to the given UTC ID.
    
    Keyword arguments:
    utcid -- UTC ID string. See utcid()
    
    """
    return '/' + utcid

def expand(text, **replacements):
    """Return the given text with all occurrences of '${key}' replaced by 'value'.
    
    Keyword arguments:
    text -- The text to expand (unmodified if None).
    **replacements -- The dictionary of replacements.
    
    Example:
    #replacing '${runid}' by '2007-06-22_13.17.51'
    expand(text, runid = '2007-06-22_13.17.51')

    """
    if text:
        for key, value in replacements.items():
            text = text.replace('${%s}' % key, value)
    return text

def writefile(path, content):
    """Write the content to the specified file, creating directories if necessary.
    
    Keyword arguments:
    path -- The path to the file (variables and tilde are expanded).
    context -- The file contexts as a string.
    
    """
    path = files.expandfilename(path)
    directory = os.path.dirname(path)
    if not os.path.exists(directory):
        os.makedirs(directory)
    f = open(path, 'w')
    try:
        f.write(content)
    finally:
        f.close()


def readfile(path):
    """Return the content of the specified file.
    
    Keyword arguments:
    path -- The path to the file (variables and tilde are expanded).
    
    """
    path = files.expandfilename(path)
    f = open(path)
    try:
        return f.read()
    finally:
        f.close()




import queue
import threading
import time
import copy
from contextlib import contextmanager

from GangaCore import GANGA_SWAN_INTEGRATION
from GangaCore.Core.GangaThread import GangaThread
from GangaCore.Core.GangaRepository import RegistryKeyError, RegistryLockError
from GangaCore.Core.exceptions import CredentialRenewalError

from GangaCore.Utility.threads import SynchronisedObject

from GangaCore.GPIDev.Credentials import credential_store, get_needed_credentials
from GangaCore.GPIDev.Credentials.AfsToken import AfsToken
from GangaCore.Core.InternalServices import Coordinator

from GangaCore.GPIDev.Base.Proxy import isType, stripProxy, getName, getRuntimeGPIObject

from GangaCore.GPIDev.Lib.Job.Job import lazyLoadJobStatus, lazyLoadJobBackend

# Setup logging ---------------
from GangaCore.Utility.logging import getLogger, log_unknown_exception, log_user_exception

from GangaCore.Core.exceptions import BackendError
from GangaCore.Utility.Config import getConfig

from collections import defaultdict

log = getLogger()

config = getConfig("PollThread")
THREAD_POOL_SIZE = config['update_thread_pool_size']
Qin = queue.Queue()
ThreadPool = []

heartbeat_times = None
global_start_time = None

# The JobAction class encapsulates a function, its arguments and its post result action
# based on what is defined as a successful run of the function.

class JobAction(object):

    __slots__ = ('function', 'args', 'kwargs', 'success', 'callback_Success', 'callback_Failure', 'thread', 'description')

    def __init__(self, function, args=(), kwargs={},
                 success=(True, ),
                 callback_Success=lambda: None,
                 callback_Failure=lambda: None):
        self.function = function
        self.args = args

        self.kwargs = kwargs
        self.success = success
        self.callback_Success = callback_Success
        self.callback_Failure = callback_Failure
        self.thread = None
        self.description = ''

def getNumAliveThreads():
    num_currently_running_command = 0
    for this_thread in ThreadPool:
        if this_thread._currently_running_command:
            num_currently_running_command += 1
    return num_currently_running_command

def checkHeartBeat(global_count):

    latest_timeNow = time.time()

    for this_thread in ThreadPool:
        thread_name = this_thread._thread_name

        last_time = heartbeat_times[thread_name]

        dead_time = config['HeartBeatTimeOut']
        max_warnings = 5

        if (latest_timeNow - last_time) > dead_time and this_thread.is_alive()\
                and this_thread._currently_running_command is True\
                and global_count < max_warnings:

            log.warning("Thread: %s Has not updated the heartbeat in %ss!! It's possibly dead" %(thread_name, str(dead_time)))
            log.warning("Thread is attempting to execute: %s" % this_thread._running_cmd)
            log.warning("With arguments: (%s)" % str(this_thread._running_args))

            ## Add at least 5sec here to avoid spamming the user non-stop that a monitoring thread has locked up almost entirely
            ## You'll get a message at most once per 5sec or less if the Monitoring is busy/asleep
            heartbeat_times[thread_name] += 5.
            global_count+=1

class MonitoringWorkerThread(GangaThread):

    __slots__ = ('_currently_running_command', '_running_cmd', '_running_args', '_thread_name')

    def __init__(self, name):
        is_critical = not config['enable_multiThreadMon']
        GangaThread.__init__(self, name, critical=is_critical)
        self._currently_running_command = False
        self._running_cmd = None
        self._running_args = None
        self._thread_name = name

    def run(self):
        self._execUpdateAction()

    # This function takes a JobAction object from the Qin queue,
    # executes the embedded function and runs post result actions.
    def _execUpdateAction(self):
        # DEBUGGING THREADS
        # import sys
        # sys.settrace(_trace)
        while not self.should_stop():
            log.debug("%s waiting..." % threading.currentThread())
            #setattr(threading.currentThread(), 'action', None)

            heartbeat_times[self._thread_name] = time.time()

            while not self.should_stop():
                try:
                    action = Qin.get(block=True, timeout=0.5)
                    break
                except queue.Empty:
                    continue

            if self.should_stop():
                break

            #setattr(threading.currentThread(), 'action', action)
            log.debug("Qin's size is currently: %d" % Qin.qsize())
            log.debug("%s running..." % threading.currentThread())
            self._currently_running_command = True
            if not isType(action, JobAction):
                continue
            if action.function == 'stop':
                break
            try:
                try:
                    self._running_cmd = action.function.__name__
                    self._running_args = []
                    for arg in action.args:
                        self._running_args.append("%s, " % arg)
                    for k, v in action.kwargs:
                        self._running_args.append("%s=%s, " % (str(k), str(v)))
                except:
                    self._running_cmd = "unknown"
                    self._running_args = []
                result = action.function(*action.args, **action.kwargs)
            except Exception as err:
                log.debug("_execUpdateAction: %s" % str(err))
                action.callback_Failure()
            else:
                if result in action.success:
                    action.callback_Success()
                else:
                    action.callback_Failure()

            self._running_args = None
            self._running_cmd = None
            self._currently_running_command = False

# Create the thread pool

def _makeThreadPool(threadPoolSize=THREAD_POOL_SIZE, daemonic=True):
    global ThreadPool, global_start_time, heartbeat_times
    global_start_time = time.time()
    if ThreadPool and len(ThreadPool) != 0:

        for this_thread in ThreadPool:
            log.error("%s running: %s" % (this_thread._thread_name, this_thread._running_cmd))

        #from GangaCore.Core.exceptions import GangaException
        #raise GangaException("Cannot doubbly init the ThreadPool! ThreadPool already populated with threads")
        log.error("Found a thread pool already in existance, wiping it and startig again!")
        del ThreadPool[:]
        ThreadPool = []

    if heartbeat_times is not None:
        heartbeat_times = None

    heartbeat_times = defaultdict( lambda: global_start_time )

    for i in range(THREAD_POOL_SIZE):
        thread_name = "MonitoringWorker_%s_%s" % (str(i), str(int(time.time()*1000)))
        t = MonitoringWorkerThread(name=thread_name)
        ThreadPool.append(t)
        t.start()


def stop_and_free_thread_pool(fail_cb=None, max_retries=5):
    """
     Clean shutdown of the thread pool. 
     A failed attempt to stop all the worker threads is followed by a call to the supplied callback which
     decides if a new attempt is performed or not.
     Example for a decision callback:      
          def myfail_cb():
             resp = raw_input("The cleanup procedures did not complete yet. Do you want to wait more?[y/N]")
             return resp.lower()=='y'
    """

    global ThreadPool

    def join_worker_threads(threads, timeout=3):
        for t in threads:
            if t.is_alive():
                t.join(timeout)
            t.stop()

    for i in range(len(ThreadPool)):
        Qin.put(JobAction('stop'))

    join_worker_threads(ThreadPool)

    # clean shutdown of stalled threads
    while True:
        if not fail_cb or max_retries <= 0:
            break
        stalled = [t for t in ThreadPool if t.is_alive()]
        if not stalled:
            break
        if fail_cb():  # continue?
            join_worker_threads(stalled, timeout=3)
            max_retries -= 1
        else:
            break

    for t in ThreadPool:
        t.stop()
    del ThreadPool[:]
    ThreadPool = []

# purge Qin

def _purge_actions_queue():
    """
    Purge Qin: consume the current queued actions 
    Note: the producer (i.e JobRegistry_Monitor) should be stopped before the method is called
    """
    # purge the queue
    for i in range(len(Qin.queue)):
        try:
            action = Qin.get_nowait()
            # let the *stop* action in queue, otherwise the worker threads will
            # fail to terminate
            if isType(action, JobAction) and action.function == 'stop':
                Qin.put(action)
        except queue.Empty:
            break

if config['autostart_monThreads'] is True:
    _makeThreadPool()


# Each entry for the updateDict_ts object (based on the UpdateDict class)
# is a _DictEntry object.
class _DictEntry(object):

    __slots__ = ('backendObj', 'jobSet', 'entryLock', 'timeoutCounterMax', 'timeoutCounter', 'timeLastUpdate')

    def __init__(self, backendObj, jobSet, entryLock, timeoutCounterMax):
        self.backendObj = backendObj
        self.jobSet = jobSet
        self.entryLock = entryLock
        self.timeoutCounterMax = timeoutCounterMax
        self.timeoutCounter = timeoutCounterMax - 0.01
        self.timeLastUpdate = 0.0

    def updateActionTuple(self):
        return self.backendObj, self.jobSet, self.entryLock


@contextmanager
def release_when_done(rlock):
    """
    A ``threading.Lock`` (or ``RLock`` etc.) object cannot be used in a context
    manager if the lock acquisition is non-blocking because the acquisition can
    fail and so the contents of the ``with`` block should not be run.

    To allow sensible ``release()`` behaviour in these cases, acquire the lock
    as usual with ``lock.acquire(False)`` and then wrap the required code with
    this context manager::

        if lock.acquire(blocking=False):
            with release_when_done(lock):
                #some code
                pass

    Attributes:
        rlock: A lock object which can have ``release()`` called on it,
    """
    try:
        yield rlock
    finally:
        rlock.release()


class UpdateDict(object):

    """
    This serves as the Update Table. Is is meant to be used 
    by wrapping it as a SynchronisedObject so as to ensure thread safety.
    """

    __slots__ = ('table',)

    def __init__(self):
        self.table = {}

    def addEntry(self, backendObj, backendCheckingFunction, jobList, timeoutMax=None):
        if not jobList:
            return
        if timeoutMax is None:
            timeoutMax = config['default_backend_poll_rate']
        #log.debug("*----addEntry()")

        backend = getName(backendObj)
        if backend in self.table:
            backendObj, jSet, lock = self.table[backend].updateActionTuple()
        else:  # New backend.
            self.table[backend] = _DictEntry(backendObj, set(jobList), threading.RLock(), timeoutMax)
            # queue to get processed
            Qin.put(JobAction(backendCheckingFunction, self.table[backend].updateActionTuple()))
            log.debug("**Adding %s to new %s backend entry." % ([stripProxy(x).getFQID('.') for x in jobList], backend))
            return True

        # backend is in Qin waiting to be processed. Increase it's list of jobs
        # by updating the table entry accordingly. This will reduce the
        # number of update requests.
        # i.e. It's like getting a friend in the queue to pay for your
        # purchases as well! ;p
        log.debug("*: backend=%s, isOwner=%s, joblist=%s, queue=%s" % (backend, lock._is_owned(), [x.id for x in jobList], Qin.qsize()))
        if lock.acquire(False):
            try:
                jSetSize = len(jSet)
                log.debug("Lock acquire successful. Updating jSet %s with %s." % ([stripProxy(x).getFQID('.') for x in jSet], [stripProxy(x).getFQID('.') for x in jobList]))
                jSet.update(jobList)
                # If jSet is empty it was cleared by an update action
                # i.e. the queue does not contain an update action for the
                # particular backend any more.
                if jSetSize:  # jSet not cleared
                    log.debug("%s backend job set exists. Added %s to it." % (backend, [stripProxy(x).getFQID('.') for x in jobList]))
                else:
                    Qin.put(JobAction(backendCheckingFunction, self.table[backend].updateActionTuple()))
                    log.debug("Added new %s backend update action for jobs %s." % (backend, [stripProxy(x).getFQID('.') for x in self.table[backend].updateActionTuple()[1]]))

            except Exception as err:
                log.error("addEntry error: %s" % str(err))
            finally:
                lock.release()

            log.debug("**: backend=%s, isOwner=%s, joblist=%s, queue=%s" %
                    (backend, lock._is_owned(), [stripProxy(x).getFQID('.') for x in jobList], Qin.qsize()))
            return True

    def clearEntry(self, backend):
        if backend in self.table:
            entry = self.table[backend]
        else:
            log.error("Error clearing the %s backend. It does not exist!" % backend)

        entry.jobSet = set()
        entry.timeoutCounter = entry.timeoutCounterMax

    @staticmethod
    def timeoutCheck(thisDict):
        for backend, entry in thisDict.table.items():
            # timeoutCounter is reset to its max value ONLY by a successful update action.
            #
            # Initial value and subsequent resets by timeoutCheck() will set the timeoutCounter
            # to a value just short of the max value to ensure that it the timeoutCounter is
            # not decremented simply because there are no updates occuring.
            if entry.timeoutCounter == entry.timeoutCounterMax and entry.entryLock.acquire(False):
                with release_when_done(entry.entryLock):
                    log.debug("%s has been reset. Acquired lock to begin countdown." % backend)
                    entry.timeLastUpdate = time.time()

                    # decrease timeout counter
                    if entry.timeoutCounter <= 0.0:
                        entry.timeoutCounter = entry.timeoutCounterMax - 0.01
                        entry.timeLastUpdate = time.time()
                        log.debug("%s backend counter timeout. Resetting to %s." % (backend, entry.timeoutCounter))
                    else:
                        _l = time.time()
                        entry.timeoutCounter -= _l - entry.timeLastUpdate
                        entry.timeLastUpdate = _l

    def isBackendLocked(self, backend):
        if backend in self.table:
            return self.table[backend].entryLock._is_owned()
        else:
            return False

    def releaseLocks(self):
        for backend, entry in self.table.items():
            if entry.entryLock._is_owned():
                entry.entryLock.release()


class CallbackHookEntry(object):

    __slots__ = ('argDict', 'enabled', 'timeout', '_lastRun')

    def __init__(self, argDict, enabled=True, timeout=0):
        self.argDict = argDict
        self.enabled = enabled
        # the frequency in seconds
        self.timeout = timeout
        # record the time when this hook has been run
        self._lastRun = 0


def autoKill_if_required(jobList_fromset):
    for j in jobList_fromset:
        if j.master:
            # Check for max number of resubmissions
            n_completed = len([sj for sj in j.master.subjobs if (sj.status == 'completed' or sj.status == 'completing')])
            n_failed = len([sj for sj in j.master.subjobs if sj.status == 'failed'])

            if n_completed == 0 and n_failed >= config['autoKillThreshold'] and not config['autoKillThreshold'] == -1:
                log.warning('Killing job %d as too many subjobs have failed with no successful completions. Please check your options!' % j.master.id)
                j.master.auto_kill()
        else:
            # Check for max number of resubmissions
            n_completed = len([sj for sj in j.subjobs if sj.status == 'completed'])
            n_failed = len([sj for sj in j.subjobs if sj.status == 'failed'])

            if n_completed == 0 and n_failed >= config['autoKillThreshold'] and not config['autoKillThreshold'] == -1:
                log.warning('Killing job %d as too many subjobs have failed with no successful completions. Please check your options!' % j.id)
                j.auto_kill()

def resubmit_if_required(jobList_fromset):
    for j in jobList_fromset:
        if not j.do_auto_resubmit:
            continue

        if len(j.subjobs) == 0:
            try_resubmit = j.info.submit_counter <= config['MaxNumResubmits']
        else:
            # Check for max number of resubmissions
            n_completed = len([sj for sj in j.subjobs if sj.status == 'completed'])
            n_failed = len([sj for sj in j.subjobs if sj.status == 'failed'])

            # Check n_failed > 0 to avoid ZeroDivisionError
            if n_failed > 0 and (float(n_failed) / float(n_completed + n_failed)) >= config['MaxFracForResubmit']:
                # Too many subjobs have failed, stop trying
                continue

            try_resubmit = n_failed > 0

        if try_resubmit and j.backend.check_auto_resubmit():
            log.warning('Auto-resubmit job %d...' % j.id)
            j.auto_resubmit()


def get_jobs_in_bunches(jobList_fromset, blocks_of_size=5, stripProxies=True):
    """
    Return a list of lists of subjobs where each list contains
    a total number of jobs close to 'blocks_of_size' as possible
    whilst not splitting jobs containing subjobs
    Also strip the jobs of their proxies...
    """
    list_of_bunches = []
    temp_list = []

    for this_job in jobList_fromset:

        if stripProxies:
            temp_list.append(stripProxy(this_job))
        else:
            temp_list.append(this_job)

        count = 0
        for found_job in temp_list:
            sj_len = len(found_job.subjobs)
            if sj_len > 0:
                count += sj_len
            else:
                count += 1

        if count >= blocks_of_size:
            list_of_bunches.append(temp_list)
            temp_list = []

    if len(temp_list) != 0:
        list_of_bunches.append(temp_list)
        temp_list = []

    return list_of_bunches


class JobRegistry_Monitor(GangaThread):
    """Job monitoring service thread."""
    
    uPollRate = 1.
    minPollRate = 1.
    global_count = 0

    __slots__ = ('registry_slice', '__sleepCounter', '__updateTimeStamp', 'progressCallback', 'callbackHookDict', 'clientCallbackDict', 'alive', 'enabled', 'steps', 'activeBackends', 'updateJobStatus', 'errors', 'updateDict_ts', '__mainLoopCond', '__cleanUpEvent', '__monStepsTerminatedEvent', 'stopIter', '_runningNow')

    def __init__(self, registry_slice):
        GangaThread.__init__(self, name="JobRegistry_Monitor")
        log.debug("Constructing JobRegistry_Monitor")
        self.setDaemon(True)
        self.registry_slice = registry_slice
        self.__sleepCounter = 0.0
        self.__updateTimeStamp = time.time()
        self.progressCallback = lambda x: None
        self.callbackHookDict = {}
        self.clientCallbackDict = {}
        self.alive = True
        self.enabled = False
        # run the monitoring loop continuosly (steps=-1) or just a specified
        # number of steps(>0)
        self.steps = -1
        self.activeBackends = {}
        self.updateJobStatus = None
        self.errors = {}

        self.updateDict_ts = SynchronisedObject(UpdateDict())

        # Create the default backend update method and add to callback hook.
        self.makeUpdateJobStatusFunction()

        # Add credential checking to monitoring loop
        for afsToken in credential_store.get_all_matching_type(AfsToken()):
            log.debug("Setting callback hook for %s" % afsToken.location)
            self.setCallbackHook(JobRegistry_Monitor.makeCredCheckJobInsertor, {'thisMonitor': self, 'credObj': afsToken}, True, timeout=config['creds_poll_rate'])

        # Add low disk-space checking to monitoring loop
        log.debug("Setting callback hook for disk space checking")
        self.setCallbackHook(JobRegistry_Monitor.diskSpaceCheckJobInsertor, {'thisMonitor': self}, True, timeout=config['diskspace_poll_rate'])

        # synch objects
        # main loop mutex
        self.__mainLoopCond = threading.Condition()
        # cleanup synch
        self.__cleanUpEvent = threading.Event()
        # asynch mon loop running synch
        self.__monStepsTerminatedEvent = threading.Event()
        # event to signal the break of job lists iterators
        self.stopIter = threading.Event()
        self.stopIter.set()

        self._runningNow = False

        if GANGA_SWAN_INTEGRATION:
            self.newly_discovered_jobs = []

    def isEnabled( self, useRunning = True ):
        if useRunning:
            return self.enabled or self.__isInProgress() and not self.steps
        else:
            return self.enabled and not self.steps

    def run(self):
        """
        Main monitoring loop
        """
        import _thread
        from GangaCore.Core.MonitoringComponent import monitoring_thread_id
        monitoring_thread_id = _thread.get_ident()
        del _thread

        log.debug("Starting run method")

        while self.alive:
            if GANGA_SWAN_INTEGRATION:
                # Monitor Jobs from other sessions
                new_jobs = stripProxy(self.registry_slice).objects.repository.update_index(True, True)
                self.newly_discovered_jobs = list(set(self.newly_discovered_jobs) | set(new_jobs))
                for i in self.newly_discovered_jobs:
                    j = stripProxy(self.registry_slice(i))
                    job_status = lazyLoadJobStatus(j)
                    if job_status in ['new']:
                        stripProxy(self.registry_slice).objects.repository.load([i])

            checkHeartBeat(JobRegistry_Monitor.global_count)
            log.debug("Monitoring Loop is alive")
            # synchronize the main loop since we can get disable requests
            with self.__mainLoopCond:

                log.debug("steps: %s" % str(self.steps))
                log.debug("%s" % str(self.registry_slice))

                log.debug("Monitoring loop __mainLoopCond")
                log.debug('Monitoring loop lock acquired. Running loop')
                # we are blocked here while the loop is disabled
                while not self.enabled and self.steps <= 0:
                    log.debug("Not enabled")
                    self.__cleanUp()
                    if not self.alive:  # stopped?
                        return
                    # disabled,but still alive, so we keep waiting
                    # for i in range( int(self.uPollRate*20) ):
                    #    if not self.alive:
                    #        return
                    #    self.__mainLoopCond.wait( self.uPollRate*0.05 )
                    self.__mainLoopCond.wait()

                log.debug("Launching Monitoring Step")
                self.__monStep()

                log.debug("Finished Step")

                # delay here the monitoring steps according to the
                # configuration
                while self.__sleepCounter > 0.0:
                    log.debug("Wait Condition")
                    self.progressCallback(self.__sleepCounter)
                    for i in range(int(self.uPollRate * 20)):
                        if self.enabled:
                            self.__mainLoopCond.wait(self.uPollRate * 0.05)
                    if not self.enabled:
                        if not self.alive:  # stopped?
                            self.__cleanUp()
                        # disabled, break to the outer while
                        break
                    else:
                        self.__sleepCounter -= self.uPollRate

                else:
                    log.debug("Run on Demand")
                    # run on demand?
                    if self.steps > 0:
                        # decrement the remaining number of steps to run
                        self.steps -= 1
                        # requested number of steps executed, disabling...
                        if self.steps <= 0:
                            self.enabled = False
                            # notify the blocking call of runMonitoring()
                            self.__monStepsTerminatedEvent.set()

        log.debug("Monitoring Cleanup")
        # final cleanup
        self.__cleanUp()

    def __monStep(self):
        """
        A single monitoring step in the monitoring loop
        Note:
        Internally the step does not block, it produces *actions* that are queued to be run 
        in the thread pool
        """
        if not self.callbackHookDict:
            log.error('No callback hooks registered')
            return

        for cbHookFunc in self.callbackHookDict.keys():

            log.debug("\n\nProcessing Function: %s" % cbHookFunc)

            if cbHookFunc in self.callbackHookDict:
                cbHookEntry = self.callbackHookDict[cbHookFunc][1]
            else:
                log.debug("Monitoring KeyError: %s" % str(cbHookFunc))
                continue

            log.debug("cbHookEntry.enabled: %s" % str(cbHookEntry.enabled))
            log.debug("(time.time() - cbHookEntry._lastRun): %s" % str((time.time() - cbHookEntry._lastRun)))
            log.debug("cbHookEntry.timeout: %s" % str(cbHookEntry.timeout))

            if cbHookEntry.enabled and (time.time() - cbHookEntry._lastRun) >= cbHookEntry.timeout:
                log.debug("Running monitoring callback hook function %s(**%s)" %(cbHookFunc, cbHookEntry.argDict))
                #self.callbackHookDict[cbHookFunc][0](**cbHookEntry.argDict)
                try:

                    if 'self' in cbHookEntry.argDict:
                        _argDict = copy.copy(cbHookEntry.argDict)
                        __self = _argDict['self']
                        del _argDict['self']
                        (self.callbackHookDict[cbHookFunc][0])(self=__self, **(_argDict))
                    else:
                        (self.callbackHookDict[cbHookFunc][0])(**(cbHookEntry.argDict))
                except Exception as err:
                    log.error("Caught Unknown Callback Exception")
                    log.error("Callback:\n %s" % str(err))
                cbHookEntry._lastRun = time.time()

        log.debug("\n\nRunning runClientCallbacks")
        self.runClientCallbacks()

        log.debug("Finished runClientCallbacks")

        self.__updateTimeStamp = time.time()
        self.__sleepCounter = config['base_poll_rate']

    
    def reloadJob(self, i):
        """
        Reload a Job from disk.
        Parameters:
        i: Job ID
        Return:
            True, if Job is successfully reloaded from disk.
        """
        stripProxy(self.registry_slice).objects.repository.load([i])
        return True

    def runMonitoring(self, jobs=None, steps=1, timeout=300):
        """
        Enable/Run the monitoring loop and wait for the monitoring steps completion.
        Parameters:
          steps:   number of monitoring steps to run
          timeout: how long to wait for monitor steps termination (seconds)
          jobs: a registry slice to be monitored (None -> all jobs), it may be passed by the user so ._impl is stripped if needed
        Return:
          False, if the loop cannot be started or the timeout occured while waiting for monitoring termination
          True, if the monitoring steps were successfully executed  
        Note:         
          This method is meant to be used in Ganga scripts to request monitoring on demand. 
        """

        log.debug("runMonitoring")

        if GANGA_SWAN_INTEGRATION:
            # Detect New Jobs from other sessions.
            new_jobs = stripProxy(self.registry_slice).objects.repository.update_index(True, True)
            self.newly_discovered_jobs = list(set(self.newly_discovered_jobs) | set(new_jobs))
            # Only load jobs from disk which are in new state currently.
            for i in self.newly_discovered_jobs:
                j = stripProxy(self.registry_slice(i))
                job_status = lazyLoadJobStatus(j)
                if job_status in ['new']:
                    stripProxy(self.registry_slice).objects.repository.load([i])

        if not isType(steps, int) and steps < 0:
            log.warning("The number of monitor steps should be a positive (non-zero) integer")
            return False

        if not self.alive:
            log.error("Cannot run the monitoring loop. It has already been stopped")
            return False

        # we don not allow the user's request the monitoring loop while the
        # internal services are stopped
        if not Coordinator.servicesEnabled:
            log.error("Cannot run the monitoring loop."
                      "The internal services are disabled (check your credentials or available disk space)")
            return False

        # if the monitoring is disabled (e.g. scripts)
        if not self.enabled:
            # and there are some required cred which are missing
            # (the monitoring loop does not monitor the credentials so we need to check 'by hand' here)
            _missingCreds = get_needed_credentials()
            if _missingCreds:
                log.error("Cannot run the monitoring loop. The following credentials are required: %s" % _missingCreds)
                return False

        #log.debug("jobs: %s" % str(jobs))
        #log.debug("self.__mainLoopCond: %s" % str(self.__mainLoopCond))

        with self.__mainLoopCond:
            log.debug('Monitoring loop lock acquired. Enabling mon loop')
            if self.enabled or self.__isInProgress():
                log.error("The monitoring loop is already running.")
                return False

            if jobs is not None:
                m_jobs = jobs

                # additional check if m_jobs is really a registry slice
                # the underlying code is not prepared to handle correctly the
                # situation if it is not
                from GangaCore.GPIDev.Lib.Registry.RegistrySlice import RegistrySlice
                if not isType(m_jobs, RegistrySlice):
                    log.warning('runMonitoring: jobs argument must be a registry slice such as a result of jobs.select() or jobs[i1:i2]')
                    return False

                #self.registry_slice = m_jobs
                #log.debug("m_jobs: %s" % str(m_jobs))
                self.makeUpdateJobStatusFunction(jobSlice=m_jobs)

            log.debug("Enable Loop, Clear Iterators and setCallbackHook")
            # enable mon loop
            self.enabled = True
            # set how many steps to run
            self.steps = steps
            # enable job list iterators
            self.stopIter.clear()
            # Start backend update timeout checking.
            self.setCallbackHook(UpdateDict.timeoutCheck, {'thisDict': self.updateDict_ts}, True)

            log.debug("Waking up Main Loop")
            # wake up the mon loop
            self.__mainLoopCond.notifyAll()

        log.debug("Waiting to execute steps")
        # wait to execute the steps
        self.__monStepsTerminatedEvent.wait()
        self.__monStepsTerminatedEvent.clear()

        log.debug("Test for timeout")
        # wait the steps to be executed or timeout to occur
        if not self.__awaitTermination(timeout):
            log.warning("Monitoring loop started but did not complete in the given timeout.")
            # force loops termination
            self.stopIter.set()
            return False
        return True

    def enableMonitoring(self):
        """
        Run the monitoring loop continuously
        """

        if not self.alive:
            log.error("Cannot start monitoring loop. It has already been stopped")
            return False

        with self.__mainLoopCond:
            log.debug('Monitoring loop lock acquired. Enabling mon loop')
            self.enabled = True
            # infinite loops
            self.steps = -1
            # enable job list iterators
            self.stopIter.clear()
            log.debug('Monitoring loop enabled')
            # Start backend update timeout checking.
            self.setCallbackHook(UpdateDict.timeoutCheck, {'thisDict': self.updateDict_ts}, True)
            self.__mainLoopCond.notifyAll()

        return True

    def disableMonitoring(self, fail_cb=None, max_retries=5):
        """
        Disable the monitoring loop
        """

        if not self.alive:
            log.error("Cannot disable monitoring loop. It has already been stopped")
            return False

        was_enabled = self.enabled

        try:
            if self.enabled:
                log.debug("Disabling Monitoring Service")

                # THIS NEEDS TO BE HERE FOR A CLEAN DISABLE OF THE MONITORING L
                self.enabled = False
                # CANNOT DETERMINE IF WE SHOULD CONTINUE WITH EXPENSIVE OUT OF
        except Exception as err:
            log.error("ERROR STOPPING MONITORING THREAD, feel free to force exit")
            # This except is left in incase we decide to add things here which
            # can fail!
            log.error("Err %s" % str(err))

        with self.__mainLoopCond:
            log.debug('Monitoring loop lock acquired. Disabling mon loop')
            self.enabled = False
            self.steps = 0
            self.stopIter.set()

            log.debug('Monitoring loop disabled')
            # wake up the monitoring loop
            self.__mainLoopCond.notifyAll()

        if was_enabled is True and self._runningNow:
            log.info("Some tasks are still running on Monitoring Loop")
            log.info("Please wait for them to finish to avoid data corruption")

        #while self._runningNow is True:
        #    time.sleep(0.5)

        #if was_enabled:
        #    log.info("Monitoring Loop has stopped")

        _purge_actions_queue()
        stop_and_free_thread_pool(fail_cb, max_retries)

        JobRegistry_Monitor.global_count = 0
        return True

    def stop(self, fail_cb=None, max_retries=5):
        """
        Stop the monitoring loop
        Parameters:
         fail_cb : if not None, this callback is called if a retry attempt is needed
        """

        if not self.alive and ThreadPool != []:
            log.warning("Monitoring loop has already been stopped")
            return False
        else:
            self.alive = False

        self.__mainLoopCond.acquire()
        if self.enabled:
            log.info('Stopping the monitoring component...')
            self.alive = False
            self.enabled = False

        try:
            # signal the main thread to finish
            self.steps = 0
            self.stopIter.set()
        except Exception as err:
            log.error("stopIter error: %s" % str(err))
        try:
            # wake up the monitoring loop
            self.__mainLoopCond.notifyAll()
        except Exception as err:
            log.error("Monitoring Stop Error: %s" % str(err))
        finally:
            self.__mainLoopCond.release()
        # wait for cleanup
        self.__cleanUpEvent.wait()
        self.__cleanUpEvent.clear()

        # ---->
        # wait for all worker threads to finish
        #self.__awaitTermination()
        # join the worker threads
        _purge_actions_queue()
        stop_and_free_thread_pool(fail_cb, max_retries)
        ###log.info( 'Monitoring component stopped successfully!' )

        #while self._runningNow is True:
        #    time.sleep(0.5)

        JobRegistry_Monitor.global_count = 0
        return True

    def __cleanUp(self):
        """
        Cleanup function ran in JobRegistry_Monitor thread to disable the monitoring loop
        updateDict_ts.timeoutCheck can hold timeout locks that need to be released 
        in order to allow the pool threads to be freed.
        """

        # cleanup the global Qin
        _purge_actions_queue()
        # release timeout check locks
        timeoutCheck = self.updateDict_ts.timeoutCheck
        if timeoutCheck in self.callbackHookDict:
            self.updateDict_ts.releaseLocks()
            self.removeCallbackHook(timeoutCheck)
        # wake up the calls waiting for cleanup
        self.__cleanUpEvent.set()

    def __isInProgress(self):
        if getNumAliveThreads() > 0:
            for this_thread in ThreadPool:
                log.debug("Thread currently running: %s" % str(this_thread._running_cmd))
        return self.steps > 0 or Qin.qsize() > 0 or getNumAliveThreads() > 0

    def __awaitTermination(self, timeout=5):
        """
         Wait for resources to be cleaned up (threads,queue)
         Returns:
             False, on timeout
        """
        while self.__isInProgress():
            time.sleep(self.uPollRate)
            timeout -= self.uPollRate
            if timeout <= 0:
                return False
        return True

    def setCallbackHook(self, func, argDict, enabled, timeout=0):
        func_name = getName(func)
        log.debug('Setting Callback hook function %s.' % func_name)
        log.debug('arg dict: %s' % str(argDict))
        #if 'self' not in argDict:
        #    argDict['self'] = self
        if func_name in self.callbackHookDict:
            log.debug('Replacing existing callback hook function %s with %s' % (str(self.callbackHookDict[func_name]), func_name))
        self.callbackHookDict[func_name] = [func, CallbackHookEntry(argDict=argDict, enabled=enabled, timeout=timeout)]

    def removeCallbackHook(self, func):
        func_name = getName(func)
        log.debug('Removing Callback hook function %s.' % func_name)
        if func_name in self.callbackHookDict:
            del self.callbackHookDict[func_name]
        else:
            log.error('Callback hook function does not exist.')

    def enableCallbackHook(self, func):
        func_name = getName(func)
        log.debug('Enabling Callback hook function %s.' % func_name)
        if func_name in self.callbackHookDict:
            self.callbackHookDict[func_name][1].enabled = True
        else:
            log.error('Callback hook function does not exist.')

    def disableCallbackHook(self, func):
        func_name = getName(func)
        log.debug('Disabling Callback hook function %s.' % func_name)
        if func_name in self.callbackHookDict:
            self.callbackHookDict[func_name][1].enabled = False
        else:
            log.error('Callback hook function does not exist.')

    def runClientCallbacks(self):
        for clientFunc in self.clientCallbackDict:
            log.debug('Running client callback hook function %s(**%s).' % (clientFunc, self.clientCallbackDict[clientFunc]))
            clientFunc(**self.clientCallbackDict[clientFunc])

    def setClientCallback(self, clientFunc, argDict):
        log.debug('Setting client callback hook function %s(**%s).' % (clientFunc, argDict))
        if clientFunc in self.clientCallbackDict:
            self.clientCallbackDict[clientFunc] = argDict
        else:
            log.error("Callback hook function not found.")

    def removeClientCallback(self, clientFunc):
        log.debug('Removing client callback hook function %s.' % clientFunc)
        if clientFunc in self.clientCallbackDict:
            del self.clientCallbackDict[clientFunc]
        else:
            log.error("%s not found in client callback dictionary." % getName(clientFunc))

    def __defaultActiveBackendsFunc(self, jobSlice=None):
        log.debug("__defaultActiveBackendsFunc")
        active_backends = {}
        
        if GANGA_SWAN_INTEGRATION:
            # Detect new Jobs from other sessions
            new_jobs = stripProxy(self.registry_slice).objects.repository.update_index(True, True)
            self.newly_discovered_jobs = list(set(self.newly_discovered_jobs) | set(new_jobs))

        # FIXME: this is not thread safe: if the new jobs are added then
        # iteration exception is raised
        if jobSlice is not None:
            fixed_ids = jobSlice.ids()
        else:
            fixed_ids = self.registry_slice.ids()
        #log.debug("Registry: %s" % str(self.registry_slice))
        log.debug("Running over fixed_ids: %s" % str(fixed_ids))
        for i in fixed_ids:
            try:
                # This is safe as it's addressing a job which _better_ be in the job repo
                j = stripProxy(self.registry_slice(i))

                job_status = lazyLoadJobStatus(j)
                
                if GANGA_SWAN_INTEGRATION:
                    # Load those Jobs from disk which are in new state to check if they are submitted
                    # in other sessions
                    if job_status in ['new']:
                        stripProxy(self.registry_slice).objects.repository.load([i])

                if job_status in ['submitted', 'running'] or (j.master and (job_status in ['submitting'])):
                    if self.enabled is True and self.alive is True:
                        backend_obj = lazyLoadJobBackend(j)
                        backend_name = getName(backend_obj)
                        active_backends.setdefault(backend_name, [])
                        active_backends[backend_name].append(j)
            except RegistryKeyError as err:
                log.debug("RegistryKeyError: The job was most likely removed")
                log.debug("RegError %s" % str(err))
            except RegistryLockError as err:
                log.debug("RegistryLockError: The job was most likely removed")
                log.debug("Reg LockError%s" % str(err))

        summary = '{'
        for backend, these_jobs in active_backends.items():
            summary += '"' + str(backend) + '" : ['
            for this_job in these_jobs:
                summary += str(stripProxy(this_job).id) + ', '#getFQID('.')) + ', '
            summary += '], '
        summary += '}'
        log.debug("Returning active_backends: %s" % summary)
        return active_backends

    # This function will be run by update threads
    def _checkBackend(self, backendObj, jobListSet, lock):

        log.debug("\n\n_checkBackend\n\n")

        currentThread = threading.currentThread()
        # timeout mechanism may have acquired the lock to impose delay.
        lock.acquire()
        self._runningNow = True

        try:
            log.debug("[Update Thread %s] Lock acquired for %s" % (currentThread, getName(backendObj)))
            #alljobList_fromset = IList(filter(lambda x: x.status in ['submitted', 'running'], jobListSet), self.stopIter)
            # print alljobList_fromset
            #masterJobList_fromset = IList(filter(lambda x: (x.master is not None) and (x.status in ['submitting']), jobListSet), self.stopIter)

            #FIXME We've lost IList and the above method for adding a job which is in a submitted state looks like one that didn't work
            # Come back and fix this once 6.1.3 is out. We can drop features for functionalist here as the lazy loading is fixed in this release
            # rcurrie
            alljobList_fromset = list([x for x in jobListSet if x.status in ['submitting', 'submitted', 'running']])
            masterJobList_fromset = list()

            # print masterJobList_fromset
            jobList_fromset = alljobList_fromset
            jobList_fromset.extend(masterJobList_fromset)
            # print jobList_fromset
            self.updateDict_ts.clearEntry(getName(backendObj))
            try:
                log.debug("[Update Thread %s] Updating %s with %s." % (currentThread, getName(backendObj), [x.id for x in jobList_fromset]))

                tested_backends = []

                for j in jobList_fromset:

                    run_setup = False

                    if backendObj is not None:
                        if hasattr(backendObj, 'setup'):
                            stripProxy(j.backend).setup()
                    else:
                        if hasattr(j.backend, 'setup'):
                            stripProxy(j.backend).setup()

                if self.enabled is False and self.alive is False:
                    log.debug("NOT enabled, leaving")
                    return

                block_size = config['numParallelJobs']
                all_job_bunches = get_jobs_in_bunches(jobList_fromset, blocks_of_size = block_size )

                bunch_size = 0
                for bunch in all_job_bunches:
                    bunch_size += len(bunch)
                assert(bunch_size == len(jobList_fromset))

                all_exceptions = []

                for this_job_list in all_job_bunches:

                    if self.enabled is False and self.alive is False:
                        log.debug("NOT enabled, breaking loop")
                        break

                    ### This tries to loop over ALL jobs in 'this_job_list' with the maximum amount of redundancy to keep
                    ### going and attempting to update all (sub)jobs if some fail
                    ### ALL ERRORS AND EXCEPTIONS ARE REPORTED VIA log.error SO NO INFORMATION IS LOST/IGNORED HERE!
                    job_ids = ''
                    for this_job in this_job_list:
                        job_ids += ' %s' % str(this_job.id) 
                    log.debug("Updating Jobs: %s" % job_ids)
                    try:
                        stripProxy(backendObj).master_updateMonitoringInformation(this_job_list)
                    except Exception as err:
                        #raise err
                        log.debug("Err: %s" % str(err))
                        ## We want to catch ALL of the exceptions
                        ## This would allow us to continue in the case of errors due to bad job/backend combinations
                        if err not in all_exceptions:
                            all_exceptions.append(err)

                if all_exceptions != []:
                    for err in all_exceptions:
                        log.error("Monitoring Error: %s" % str(err))
                    ## We should be raising exceptions no matter what
                    raise all_exceptions[0]

                autoKill_if_required(jobList_fromset)
                resubmit_if_required(jobList_fromset)

            except BackendError as x:
                self._handleError(x, x.backend_name, 0)
            except Exception as err:
                #self._handleError(err, getName(backendObj), 1)
                log.error("Monitoring Error: %s" % str(err))
                log.debug("Lets not crash here!")
                return

            # FIXME THIS METHOD DOES NOT EXIST
            #log.debug("[Update Thread %s] Flushing registry %s." % (currentThread, [x.id for x in jobList_fromset]))
            # FIXME THIS RETURNS A REGISTRYSLICE OBJECT NOT A REGISTRY, IS THIS CORRECT? SHOULD WE FLUSH
            # COMMENTING OUT AS IT SIMPLY WILL NOT RUN/RESOLVE!
            # this concerns me - rcurrie
            #self.registry_slice._flush(jobList_fromset)  # Optimisation required!

            #for this_job in jobList_fromset:
            #    stripped_job = stripProxy(this_job)
            #    stripped_job._getRegistry()._flush([stripped_job])

        except Exception as err:
            log.debug("Monitoring Loop Error: %s" % str(err))
        finally:
            lock.release()
            log.debug("[Update Thread %s] Lock released for %s." % (currentThread, getName(backendObj)))
            self._runningNow = False

        log.debug("Finishing _checkBackend")
        return

    @staticmethod
    def _checkActiveBackends(thisMonitor, activeBackendsFunc, jobSlice):

        log.debug("calling function _checkActiveBackends")
        if jobSlice is None:
            activeBackends = activeBackendsFunc()
        else:
            activeBackends = activeBackendsFunc(jobSlice=jobSlice)

        summary = '{'
        for this_backend, these_jobs in activeBackends.items():
            summary += '"' + this_backend + '" : ['
            for this_job in these_jobs:
                summary += str(stripProxy(this_job).getFQID('.')) + ', '
            summary += '], '
        summary += '}'
        log.debug("Active Backends: %s" % summary)

        for jList in activeBackends.values():

            #log.debug("backend: %s" % str(jList))
            backendObj = jList[0].backend
            b_name = getName(backendObj)
            if b_name in config:
                pRate = config[b_name]
            else:
                pRate = config['default_backend_poll_rate']

            # TODO: To include an if statement before adding entry to
            #       updateDict. Entry is added only if credential requirements
            #       of the particular backend is satisfied.
            #       This requires backends to hold relevant information on its
            #       credential requirements.
            #log.debug("addEntry: %s, %s, %s, %s" % (str(backendObj), str(thisMonitor._checkBackend), str(jList), str(pRate)))
            thisMonitor.updateDict_ts.addEntry(backendObj, thisMonitor._checkBackend, jList, pRate)
            summary = str([stripProxy(x).getFQID('.') for x in jList])
            log.debug("jList: %s" % str(summary))


    def makeUpdateJobStatusFunction(self, makeActiveBackendsFunc=None, jobSlice=None):
        log.debug("makeUpdateJobStatusFunction")
        if makeActiveBackendsFunc is None:
            makeActiveBackendsFunc = self.__defaultActiveBackendsFunc

        if self.updateJobStatus is not None:
            self.removeCallbackHook(self.updateJobStatus)
        self.updateJobStatus = self._checkActiveBackends
        self.setCallbackHook(self._checkActiveBackends, {'thisMonitor': self, 'activeBackendsFunc': makeActiveBackendsFunc, 'jobSlice': jobSlice}, True)
        log.debug("Returning")
        return self.updateJobStatus

    @staticmethod
    def makeCredCheckJobInsertor(thisMonitor, credObj):
        def credCheckJobInsertor():
            def cb_Success():
                thisMonitor.enableCallbackHook(credCheckJobInsertor)

            def cb_Failure():
                thisMonitor.enableCallbackHook(credCheckJobInsertor)
                thisMonitor._handleError('%s checking failed!' % getName(credObj), getName(credObj), False)

            log.debug('Inserting %s checking function to Qin.' % getName(credObj))
            _action = JobAction(function=thisMonitor.makeCredChecker(credObj),
                                callback_Success=cb_Success,
                                callback_Failure=cb_Failure)
            thisMonitor.disableCallbackHook(credCheckJobInsertor)
            try:
                Qin.put(_action)
            except Exception as err:
               log.debug("makeCred Err: %s" % str(err))
               cb_Failure("Put _action failure: %s" % str(_action), "unknown", True )
        return credCheckJobInsertor

    def makeCredChecker(self, credObj):
        def credChecker():
            log.debug("Checking %s." % getName(credObj))
            try:
                credObj.renew()
            except CredentialRenewalError:
                return False
            except Exception as err:
                logger.error("Unexpected exception!")
                logger.error("Error: %s" % err)
                return False
            else:
                return True
        return credChecker

    @staticmethod
    def diskSpaceCheckJobInsertor(thisMonitor):
        """
        Inserts the disk space checking task in the monitoring task queue
        """
        def cb_Success():
            thisMonitor.enableCallbackHook(thisMonitor.diskSpaceCheckJobInsertor)

        def cb_Failure():
            thisMonitor.disableCallbackHook(thisMonitor.diskSpaceCheckJobInsertor)
            thisMonitor._handleError('Available disk space checking failed and it has been disabled!', 'DiskSpaceChecker', False)

        log.debug('Inserting disk space checking function to Qin.')
        _action = JobAction(function=Coordinator._diskSpaceChecker,
                            callback_Success=cb_Success,
                            callback_Failure=cb_Failure)
        thisMonitor.disableCallbackHook(thisMonitor.diskSpaceCheckJobInsertor)
        try:
            Qin.put(_action)
        except Exception as err:
            log.debug("diskSp Err: %s" % str(err))
            cb_Failure()

    def updateJobs(self):
        if time.time() - self.__updateTimeStamp >= self.minPollRate:
            self.__sleepCounter = 0.0
        else:
            self.progressCallback("Processing... Please wait.")
            log.debug("Updates too close together... skipping latest update request.")
            self.__sleepCounter = self.minPollRate

    def _handleError(self, x, backend_name, show_traceback):
        def log_error():
            log.error('Problem in the monitoring loop: %s', str(x))
            #if show_traceback:
            #    log.error("exception: ", exc_info=1)
            #    #log_unknown_exception()
            #    import traceback
            #    traceback.print_stack()
            if show_traceback:
                log_user_exception(log)
        bn = backend_name
        self.errors.setdefault(bn, 0)
        if self.errors[bn] == 0:
            log_error()
            if not config['repeat_messages']:
                log.info('Further error messages from %s handler in the monitoring loop will be skipped.' % bn)
        else:
            if config['repeat_messages']:
                log_error()
        self.errors[bn] += 1


######## THREAD POOL DEBUGGING ###########
def _trace(frame, event, arg):
    setattr(threading.currentThread(), '_frame', frame)


def getStackTrace():
    import inspect

    try:
        status = "Available threads:\n"

        for worker in ThreadPool:
            status = status + "  " + worker.getName() + ":\n"

            if hasattr(worker, '_frame'):
                frame = worker._frame
                if frame:
                    status = status + "    stack:\n"
                    for frame, filename, line, function_name, context, index in inspect.getouterframes(frame):
                        status = status + "      " + function_name + " @ " + filename + " # " + str(line) + "\n"

            status = status + "\n"
        ## CANNOT CONVERT TO A STRING!!!
        #log.info("Queue", str(Qin.queue))
        log.debug("Trace: %s" % str(status))
        return status
    except Exception as err:
        print("Err: %s" % str(err))
    finally:
        pass

